"""The |FHMRUVV| |N3LO| Altarelli-Parisi splitting kernels approximations.

Authors follow Pegasus convention and so there is an additional global
minus sign with respect to our conventions.
"""

import numba as nb
import numpy as np

from .ggg import gamma_gg
from .ggq import gamma_gq
from .gnsm import gamma_nsm
from .gnsp import gamma_nsp
from .gnsv import gamma_nsv
from .gps import gamma_ps
from .gqg import gamma_qg


@nb.njit(cache=True)
def gamma_singlet(N, nf, cache, variation):
    r"""Compute the |N3LO| singlet anomalous dimension matrix.

      .. math::
          \gamma_S^{(3)} = \left(\begin{array}{cc}
          \gamma_{qq}^{(3)} & \gamma_{qg}^{(3)}\\
          \gamma_{gq}^{(3)} & \gamma_{gg}^{(3)}
          \end{array}\right)

    Parameters
    ----------
    N : complex
      Mellin moment
    nf : int
      Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache
    variation : tuple
        |N3LO| anomalous dimension variation ``(gg, gq, qg, qq)``

    Returns
    -------
    numpy.ndarray
        |N3LO| singlet anomalous dimension matrix
        :math:`\gamma_{S}^{(3)}(N)`
    """
    gamma_qq = gamma_nsp(N, nf, cache, variation[3]) + gamma_ps(
        N, nf, cache, variation[3]
    )
    gamma_S_0 = np.array(
        [
            [gamma_qq, gamma_qg(N, nf, cache, variation[2])],
            [
                gamma_gq(N, nf, cache, variation[1]),
                gamma_gg(N, nf, cache, variation[0]),
            ],
        ],
        np.complex_,
    )
    return gamma_S_0


@nb.njit(cache=True)
def gamma_singlet_qed(N, nf, cache, variation):
    r"""Compute the leading-order singlet anomalous dimension matrix for the
    unified evolution basis.

    .. math::
        \\gamma_S^{(3,0)} = \\left(\begin{array}{cccc}
        \\gamma_{gg}^{(3,0)} & 0 & \\gamma_{gq}^{(3,0)} & 0\\
        0 & 0 & 0 & 0 \\
        \\gamma_{qg}^{(3,0)} & 0 & \\gamma_{qq}^{(3,0)} & 0 \\
        0 & 0 & 0 & \\gamma_{qq}^{(3,0)} \\
        \\end{array}\right)

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache
    variation : tuple
        |N3LO| anomalous dimension variation ``(gg, gq, qg, qq)``

    Returns
    -------
    numpy.ndarray
        Leading-order singlet anomalous dimension matrix :math:`\\gamma_{S}^{(3,0)}(N)`
    """
    gamma_np_p = gamma_nsp(N, nf, cache, variation[3])
    gamma_qq = gamma_np_p + gamma_ps(N, nf, cache, variation[3])
    gamma_S = np.array(
        [
            [
                gamma_gg(N, nf, cache, variation[0]),
                0.0 + 0.0j,
                gamma_gq(N, nf, cache, variation[1]),
                0.0 + 0.0j,
            ],
            [0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j],
            [gamma_qg(N, nf, cache, variation[2]), 0.0 + 0.0j, gamma_qq, 0.0 + 0.0j],
            [0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j, gamma_np_p],
        ],
        np.complex_,
    )
    return gamma_S


@nb.njit(cache=True)
def gamma_valence_qed(N, nf, cache, variation):
    r"""Compute the leading-order valence anomalous dimension matrix for the
    unified evolution basis.

    .. math::
        \\gamma_V^{(3,0)} = \\left(\begin{array}{cc}
        \\gamma_{nsV}^{(3,0)} & 0\\
        0 & \\gamma_{ns-}^{(3,0)}
        \\end{array}\right)

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache
    variation : tuple
        |N3LO| anomalous dimension variation ``(nsm, nsv)``

    Returns
    -------
    numpy.ndarray
        Leading-order singlet anomalous dimension matrix :math:`\\gamma_{V}^{(3,0)}(N)`
    """
    gamma_V = np.array(
        [
            [gamma_nsv(N, nf, cache, variation[-1]), 0.0],
            [0.0, gamma_nsm(N, nf, cache, variation[-2])],
        ],
        np.complex_,
    )
    return gamma_V



r"""Contains the central operator classes.

See :doc:`Operator overview </code/Operators>`.
"""

import functools
import logging
import os
import time
from dataclasses import dataclass
from multiprocessing import Pool
from typing import Dict, Tuple

import numba as nb
import numpy as np
from scipy import integrate

import ekore.anomalous_dimensions.polarized.space_like as ad_ps
import ekore.anomalous_dimensions.unpolarized.space_like as ad_us
import ekore.anomalous_dimensions.unpolarized.time_like as ad_ut

from .. import basis_rotation as br
from .. import interpolation, mellin
from .. import scale_variations as sv
from ..couplings import Couplings
from ..interpolation import InterpolatorDispatcher
from ..io.types import EvolutionMethod, OperatorLabel
from ..kernels import ev_method
from ..kernels import non_singlet as ns
from ..kernels import non_singlet_qed as qed_ns
from ..kernels import singlet as s
from ..kernels import singlet_qed as qed_s
from ..kernels import valence_qed as qed_v
from ..matchings import Atlas, Segment, lepton_number
from ..member import OpMember
from ..scale_variations import expanded as sv_expanded
from ..scale_variations import exponentiated as sv_exponentiated

logger = logging.getLogger(__name__)


@nb.njit(cache=True)
def select_singlet_element(ker, mode0, mode1):
    """Select element of the singlet matrix.

    Parameters
    ----------
    ker : numpy.ndarray
        singlet integration kernel
    mode0 : int
        id for first sector element
    mode1 : int
        id for second sector element

    Returns
    -------
    complex
        singlet integration kernel element
    """
    j = 0 if mode0 == 100 else 1
    k = 0 if mode1 == 100 else 1
    return ker[j, k]


@nb.njit(cache=True)
def select_QEDsinglet_element(ker, mode0, mode1):
    """Select element of the QEDsinglet matrix.

    Parameters
    ----------
    ker : numpy.ndarray
        QEDsinglet integration kernel
    mode0 : int
        id for first sector element
    mode1 : int
        id for second sector element
    Returns
    -------
    ker : complex
        QEDsinglet integration kernel element
    """
    if mode0 == 21:
        index1 = 0
    elif mode0 == 22:
        index1 = 1
    elif mode0 == 100:
        index1 = 2
    else:
        index1 = 3
    if mode1 == 21:
        index2 = 0
    elif mode1 == 22:
        index2 = 1
    elif mode1 == 100:
        index2 = 2
    else:
        index2 = 3
    return ker[index1, index2]


@nb.njit(cache=True)
def select_QEDvalence_element(ker, mode0, mode1):
    """Select element of the QEDvalence matrix.

    Parameters
    ----------
    ker : numpy.ndarray
        QEDvalence integration kernel
    mode0 : int
        id for first sector element
    mode1 : int
        id for second sector element
    Returns
    -------
    ker : complex
        QEDvalence integration kernel element
    """
    index1 = 0 if mode0 == 10200 else 1
    index2 = 0 if mode1 == 10200 else 1
    return ker[index1, index2]


spec = [
    ("is_singlet", nb.boolean),
    ("is_QEDsinglet", nb.boolean),
    ("is_QEDvalence", nb.boolean),
    ("is_log", nb.boolean),
    ("logx", nb.float64),
    ("u", nb.float64),
]


@nb.experimental.jitclass(spec)
class QuadKerBase:
    """Manage the common part of Mellin inversion integral.

    Parameters
    ----------
    u : float
        quad argument
    is_log : boolean
        is a logarithmic interpolation
    logx : float
        Mellin inversion point
    mode0 : str
        first sector element
    """

    def __init__(self, u, is_log, logx, mode0):
        self.is_singlet = mode0 in [100, 21, 90]
        self.is_QEDsinglet = mode0 in [21, 22, 100, 101, 90]
        self.is_QEDvalence = mode0 in [10200, 10204]
        self.is_log = is_log
        self.u = u
        self.logx = logx

    @property
    def path(self):
        """Return the associated instance of :class:`eko.mellin.Path`."""
        if self.is_singlet or self.is_QEDsinglet:
            return mellin.Path(self.u, self.logx, True)
        else:
            return mellin.Path(self.u, self.logx, False)

    @property
    def n(self):
        """Returs the Mellin moment :math:`N`."""
        return self.path.n

    def integrand(
        self,
        areas,
    ):
        """Get transformation to Mellin space integral.

        Parameters
        ----------
        areas : tuple
            basis function configuration

        Returns
        -------
        complex
            common mellin inversion integrand
        """
        if self.logx == 0.0:
            return 0.0
        pj = interpolation.evaluate_grid(self.path.n, self.is_log, self.logx, areas)
        if pj == 0.0:
            return 0.0
        return self.path.prefactor * pj * self.path.jac


@nb.njit(cache=True)
def quad_ker(
    u,
    order,
    mode0,
    mode1,
    method,
    is_log,
    logx,
    areas,
    as_list,
    mu2_from,
    mu2_to,
    a_half,
    alphaem_running,
    nf,
    L,
    ev_op_iterations,
    ev_op_max_order,
    sv_mode,
    is_threshold,
    n3lo_ad_variation,
    is_polarized,
    is_time_like,
    use_fhmruvv,
):
    """Raw evolution kernel inside quad.

    Parameters
    ----------
    u : float
        quad argument
    order : int
        perturbation order
    mode0: int
        pid for first sector element
    mode1 : int
        pid for second sector element
    method : str
        method
    is_log : boolean
        is a logarithmic interpolation
    logx : float
        Mellin inversion point
    areas : tuple
        basis function configuration
    as1 : float
        target coupling value
    as0 : float
        initial coupling value
    mu2_from : float
        initial value of mu2
    mu2_from : float
        final value of mu2
    aem_list : list
        list of electromagnetic coupling values
    alphaem_running : bool
        whether alphaem is running or not
    nf : int
        number of active flavors
    L : float
        logarithm of the squared ratio of factorization and renormalization scale
    ev_op_iterations : int
        number of evolution steps
    ev_op_max_order : int
        perturbative expansion order of U
    sv_mode: int, `enum.IntEnum`
        scale variation mode, see `eko.scale_variations.Modes`
    is_threshold : boolean
        is this an intermediate threshold operator?
    n3lo_ad_variation : tuple
        |N3LO| anomalous dimension variation ``(gg, gq, qg, qq, nsp, nsm, nsv)``
    is_polarized : boolean
        is polarized evolution ?
    is_time_like : boolean
        is time-like evolution ?
    use_fhmruvv : bool
        if True use the |FHMRUVV| |N3LO| anomalous dimension

    Returns
    -------
    float
        evaluated integration kernel
    """
    ker_base = QuadKerBase(u, is_log, logx, mode0)
    integrand = ker_base.integrand(areas)
    if integrand == 0.0:
        return 0.0
    if order[1] == 0:
        ker = quad_ker_qcd(
            ker_base,
            order,
            mode0,
            mode1,
            method,
            as_list[-1],
            as_list[0],
            nf,
            L,
            ev_op_iterations,
            ev_op_max_order,
            sv_mode,
            is_threshold,
            is_polarized,
            is_time_like,
            n3lo_ad_variation,
            use_fhmruvv,
        )
    else:
        ker = quad_ker_qed(
            ker_base,
            order,
            mode0,
            mode1,
            method,
            as_list,
            mu2_from,
            mu2_to,
            a_half,
            alphaem_running,
            nf,
            L,
            ev_op_iterations,
            ev_op_max_order,
            sv_mode,
            is_threshold,
            n3lo_ad_variation,
            use_fhmruvv,
        )

    # recombine everything
    return np.real(ker * integrand)


@nb.njit(cache=True)
def quad_ker_qcd(
    ker_base,
    order,
    mode0,
    mode1,
    method,
    as1,
    as0,
    nf,
    L,
    ev_op_iterations,
    ev_op_max_order,
    sv_mode,
    is_threshold,
    is_polarized,
    is_time_like,
    n3lo_ad_variation,
    use_fhmruvv,
):
    """Raw evolution kernel inside quad.

    Parameters
    ----------
    quad_ker : float
        quad argument
    order : int
        perturbation order
    mode0: int
        pid for first sector element
    mode1 : int
        pid for second sector element
    method : str
        method
    as1 : float
        target coupling value
    as0 : float
        initial coupling value
    nf : int
        number of active flavors
    L : float
        logarithm of the squared ratio of factorization and renormalization scale
    ev_op_iterations : int
        number of evolution steps
    ev_op_max_order : int
        perturbative expansion order of U
    sv_mode: int, `enum.IntEnum`
        scale variation mode, see `eko.scale_variations.Modes`
    is_threshold : boolean
        is this an itermediate threshold operator?
    n3lo_ad_variation : tuple
        |N3LO| anomalous dimension variation ``(gg, gq, qg, qq, nsp, nsm, nsv)``
    use_fhmruvv : bool
        if True use the |FHMRUVV| |N3LO| anomalous dimensions

    Returns
    -------
    float
        evaluated integration kernel
    """
    # compute the actual evolution kernel for pure QCD
    if ker_base.is_singlet:
        if is_polarized:
            if is_time_like:
                raise NotImplementedError("Polarized, time-like is not implemented")
            else:
                gamma_singlet = ad_ps.gamma_singlet(order, ker_base.n, nf)
        else:
            if is_time_like:
                gamma_singlet = ad_ut.gamma_singlet(order, ker_base.n, nf)
            else:
                gamma_singlet = ad_us.gamma_singlet(
                    order, ker_base.n, nf, n3lo_ad_variation, use_fhmruvv
                )
        # scale var exponentiated is directly applied on gamma
        if sv_mode == sv.Modes.exponentiated:
            gamma_singlet = sv_exponentiated.gamma_variation(
                gamma_singlet, order, nf, L
            )
        ker = s.dispatcher(
            order,
            method,
            gamma_singlet,
            as1,
            as0,
            nf,
            ev_op_iterations,
            ev_op_max_order,
        )
        # scale var expanded is applied on the kernel
        if sv_mode == sv.Modes.expanded and not is_threshold:
            ker = np.ascontiguousarray(
                sv_expanded.singlet_variation(gamma_singlet, as1, order, nf, L, dim=2)
            ) @ np.ascontiguousarray(ker)
        ker = select_singlet_element(ker, mode0, mode1)
    else:
        if is_polarized:
            if is_time_like:
                raise NotImplementedError("Polarized, time-like is not implemented")
            else:
                gamma_ns = ad_ps.gamma_ns(order, mode0, ker_base.n, nf)
        else:
            if is_time_like:
                gamma_ns = ad_ut.gamma_ns(order, mode0, ker_base.n, nf)
            else:
                gamma_ns = ad_us.gamma_ns(
                    order, mode0, ker_base.n, nf, n3lo_ad_variation, use_fhmruvv
                )
        if sv_mode == sv.Modes.exponentiated:
            gamma_ns = sv_exponentiated.gamma_variation(gamma_ns, order, nf, L)
        ker = ns.dispatcher(
            order,
            method,
            gamma_ns,
            as1,
            as0,
            nf,
            ev_op_iterations,
        )
        if sv_mode == sv.Modes.expanded and not is_threshold:
            ker = sv_expanded.non_singlet_variation(gamma_ns, as1, order, nf, L) * ker
    return ker


@nb.njit(cache=True)
def quad_ker_qed(
    ker_base,
    order,
    mode0,
    mode1,
    method,
    as_list,
    mu2_from,
    mu2_to,
    a_half,
    alphaem_running,
    nf,
    L,
    ev_op_iterations,
    ev_op_max_order,
    sv_mode,
    is_threshold,
    n3lo_ad_variation,
    use_fhmruvv,
):
    """Raw evolution kernel inside quad.

    Parameters
    ----------
    ker_base : QuadKerBase
        quad argument
    order : int
        perturbation order
    mode0: int
        pid for first sector element
    mode1 : int
        pid for second sector element
    method : str
        method
    as1 : float
        target coupling value
    as0 : float
        initial coupling value
    mu2_from : float
        initial value of mu2
    mu2_from : float
        final value of mu2
    aem_list : list
        list of electromagnetic coupling values
    alphaem_running : bool
        whether alphaem is running or not
    nf : int
        number of active flavors
    L : float
        logarithm of the squared ratio of factorization and renormalization scale
    ev_op_iterations : int
        number of evolution steps
    ev_op_max_order : int
        perturbative expansion order of U
    sv_mode: int, `enum.IntEnum`
        scale variation mode, see `eko.scale_variations.Modes`
    is_threshold : boolean
        is this an itermediate threshold operator?
    n3lo_ad_variation : tuple
        |N3LO| anomalous dimension variation ``(gg, gq, qg, qq, nsp, nsm, nsv)``
    use_fhmruvv : bool
        if True use the |FHMRUVV| |N3LO| anomalous dimensions

    Returns
    -------
    float
        evaluated integration kernel
    """
    # compute the actual evolution kernel for QEDxQCD
    if ker_base.is_QEDsinglet:
        gamma_s = ad_us.gamma_singlet_qed(
            order, ker_base.n, nf, n3lo_ad_variation, use_fhmruvv
        )
        # scale var exponentiated is directly applied on gamma
        if sv_mode == sv.Modes.exponentiated:
            gamma_s = sv_exponentiated.gamma_variation_qed(
                gamma_s, order, nf, lepton_number(mu2_to), L, alphaem_running
            )
        ker = qed_s.dispatcher(
            order,
            method,
            gamma_s,
            as_list,
            a_half,
            nf,
            ev_op_iterations,
            ev_op_max_order,
        )
        # scale var expanded is applied on the kernel
        # TODO : in this way a_half[-1][1] is the aem value computed in
        # the middle point of the last step. Instead we want aem computed in mu2_final.
        # However the distance between the two is very small and affects only the running aem
        if sv_mode == sv.Modes.expanded and not is_threshold:
            ker = np.ascontiguousarray(
                sv_expanded.singlet_variation_qed(
                    gamma_s, as_list[-1], a_half[-1][1], alphaem_running, order, nf, L
                )
            ) @ np.ascontiguousarray(ker)
        ker = select_QEDsinglet_element(ker, mode0, mode1)
    elif ker_base.is_QEDvalence:
        gamma_v = ad_us.gamma_valence_qed(
            order, ker_base.n, nf, n3lo_ad_variation, use_fhmruvv
        )
        # scale var exponentiated is directly applied on gamma
        if sv_mode == sv.Modes.exponentiated:
            gamma_v = sv_exponentiated.gamma_variation_qed(
                gamma_v, order, nf, lepton_number(mu2_to), L, alphaem_running
            )
        ker = qed_v.dispatcher(
            order,
            method,
            gamma_v,
            as_list,
            a_half,
            nf,
            ev_op_iterations,
            ev_op_max_order,
        )
        # scale var expanded is applied on the kernel
        if sv_mode == sv.Modes.expanded and not is_threshold:
            ker = np.ascontiguousarray(
                sv_expanded.valence_variation_qed(
                    gamma_v, as_list[-1], a_half[-1][1], alphaem_running, order, nf, L
                )
            ) @ np.ascontiguousarray(ker)
        ker = select_QEDvalence_element(ker, mode0, mode1)
    else:
        gamma_ns = ad_us.gamma_ns_qed(
            order, mode0, ker_base.n, nf, n3lo_ad_variation, use_fhmruvv
        )
        # scale var exponentiated is directly applied on gamma
        if sv_mode == sv.Modes.exponentiated:
            gamma_ns = sv_exponentiated.gamma_variation_qed(
                gamma_ns, order, nf, lepton_number(mu2_to), L, alphaem_running
            )
        ker = qed_ns.dispatcher(
            order,
            method,
            gamma_ns,
            as_list,
            a_half[:, 1],
            alphaem_running,
            nf,
            ev_op_iterations,
            mu2_from,
            mu2_to,
        )
        if sv_mode == sv.Modes.expanded and not is_threshold:
            ker = (
                sv_expanded.non_singlet_variation_qed(
                    gamma_ns, as_list[-1], a_half[-1][1], alphaem_running, order, nf, L
                )
                * ker
            )
    return ker


OpMembers = Dict[OperatorLabel, OpMember]
"""Map of all operators."""


@dataclass(frozen=True)
class Managers:
    """Set of steering objects."""

    atlas: Atlas
    couplings: Couplings
    interpolator: InterpolatorDispatcher


class Operator(sv.ScaleVariationModeMixin):
    """Internal representation of a single EKO.

    The actual matrices are computed upon calling :meth:`compute`.

    Parameters
    ----------
    config : dict
        configuration
    managers : dict
        managers
    nf : int
        number of active flavors
    q2_from : float
        evolution source
    q2_to : float
        evolution target
    mellin_cut : float
        cut to the upper limit in the mellin inversion
    is_threshold : bool
        is this an itermediate threshold operator?
    """

    log_label = "Evolution"
    # complete list of possible evolution operators labels
    full_labels: Tuple[OperatorLabel, ...] = br.full_labels
    full_labels_qed: Tuple[OperatorLabel, ...] = br.full_unified_labels

    def __init__(
        self,
        config,
        managers: Managers,
        segment: Segment,
        mellin_cut=5e-2,
        is_threshold=False,
    ):
        self.config = config
        self.managers = managers
        self.nf = segment.nf
        self.q2_from = segment.origin
        self.q2_to = segment.target
        # TODO make 'cut' external parameter?
        self._mellin_cut = mellin_cut
        self.is_threshold = is_threshold
        self.op_members: OpMembers = {}
        self.order = tuple(config["order"])
        self.alphaem_running = self.managers.couplings.alphaem_running
        if self.log_label == "Evolution":
            self.a = self.compute_a()
            self.as_list, self.a_half_list = self.compute_aem_list()

    @property
    def n_pools(self):
        """Return number of parallel cores."""
        n_pools = self.config["n_integration_cores"]
        if n_pools > 0:
            return n_pools
        # so we subtract from the maximum number
        return max(os.cpu_count() + n_pools, 1)

    @property
    def xif2(self):
        r"""Return scale variation factor :math:`(\mu_F/\mu_R)^2`."""
        return self.config["xif2"]

    @property
    def int_disp(self) -> InterpolatorDispatcher:
        """Return the interpolation dispatcher."""
        return self.managers.interpolator

    @property
    def grid_size(self):
        """Return the grid size."""
        return self.int_disp.xgrid.size

    @property
    def mu2(self):
        """Return the arguments to the :math:`a_s` function."""
        mu0 = self.q2_from * (
            self.xif2 if self.sv_mode == sv.Modes.exponentiated else 1.0
        )
        mu1 = self.q2_to * (
            self.xif2
            if (not self.is_threshold and self.sv_mode == sv.Modes.expanded)
            or self.sv_mode == sv.Modes.exponentiated
            else 1.0
        )
        return (mu0, mu1)

    def compute_a(self):
        """Return the computed values for :math:`a_s` and :math:`a_{em}`."""
        coupling = self.managers.couplings
        a0 = coupling.a(
            self.mu2[0],
            nf_to=self.nf,
        )
        a1 = coupling.a(
            self.mu2[1],
            nf_to=self.nf,
        )
        return (a0, a1)

    @property
    def a_s(self):
        """Return the computed values for :math:`a_s`."""
        return (self.a[0][0], self.a[1][0])

    @property
    def a_em(self):
        """Return the computed values for :math:`a_{em}`."""
        return (self.a[0][1], self.a[1][1])

    def compute_aem_list(self):
        """Return the list of the couplings for the different values of
        :math:`a_s`.

        This functions is needed in order to compute the values of :math:`a_s`
        and :math:`a_em` in the middle point of the :math:`mu^2` interval, and
        the values of :math:`a_s` at the borders of every intervals.
        This is needed in the running_alphaem solution.
        """
        ev_op_iterations = self.config["ev_op_iterations"]
        if self.order[1] == 0:
            as_list = np.array([self.a_s[0], self.a_s[1]])
            a_half = np.zeros((ev_op_iterations, 2))
        else:
            couplings = self.managers.couplings
            mu2_steps = np.geomspace(self.q2_from, self.q2_to, 1 + ev_op_iterations)
            mu2_l = mu2_steps[0]
            as_list = np.array(
                [couplings.a_s(scale_to=mu2, nf_to=self.nf) for mu2 in mu2_steps]
            )
            a_half = np.zeros((ev_op_iterations, 2))
            for step, mu2_h in enumerate(mu2_steps[1:]):
                mu2_half = (mu2_h + mu2_l) / 2.0
                a_s, aem = couplings.a(scale_to=mu2_half, nf_to=self.nf)
                a_half[step] = [a_s, aem]
                mu2_l = mu2_h
        return as_list, a_half

    @property
    def labels(self):
        """Compute necessary sector labels to compute.

        Returns
        -------
        list(str)
            sector labels
        """
        labels = []
        # the NS sector is dynamic
        if self.order[1] == 0:
            if self.config["debug_skip_non_singlet"]:
                logger.warning("%s: skipping non-singlet sector", self.log_label)
            else:
                # add + as default
                labels.append(br.non_singlet_labels[1])
                if self.order[0] >= 2:  # - becomes different starting from NLO
                    labels.append(br.non_singlet_labels[0])
                if self.order[0] >= 3:  # v also becomes different starting from NNLO
                    labels.append(br.non_singlet_labels[2])
            # singlet sector is fixed
            if self.config["debug_skip_singlet"]:
                logger.warning("%s: skipping singlet sector", self.log_label)
            else:
                labels.extend(br.singlet_labels)
        else:
            if self.config["debug_skip_non_singlet"]:
                logger.warning("%s: skipping non-singlet sector", self.log_label)
            else:
                # add +u and +d as default
                labels.append(br.non_singlet_unified_labels[0])
                labels.append(br.non_singlet_unified_labels[2])
                # -u and -d become different starting from O(as1aem1) or O(aem2)
                # but at this point order is at least (1, 1)
                labels.append(br.non_singlet_unified_labels[1])
                labels.append(br.non_singlet_unified_labels[3])
                labels.extend(br.valence_unified_labels)
            if self.config["debug_skip_singlet"]:
                logger.warning("%s: skipping singlet sector", self.log_label)
            else:
                labels.extend(br.singlet_unified_labels)
        return labels

    @property
    def ev_method(self):
        """Return the evolution method."""
        return ev_method(EvolutionMethod(self.config["method"]))

    def quad_ker(self, label, logx, areas):
        """Return partially initialized integrand function.

        Parameters
        ----------
        label: tuple
            operator element pids
        logx: float
            Mellin inversion point
        areas : tuple
            basis function configuration

        Returns
        -------
        functools.partial
            partially initialized integration kernel
        """
        return functools.partial(
            quad_ker,
            order=self.order,
            mode0=label[0],
            mode1=label[1],
            method=self.ev_method,
            is_log=self.int_disp.log,
            logx=logx,
            areas=areas,
            as_list=self.as_list,
            mu2_from=self.q2_from,
            mu2_to=self.q2_to,
            a_half=self.a_half_list,
            alphaem_running=self.alphaem_running,
            nf=self.nf,
            L=np.log(self.xif2),
            ev_op_iterations=self.config["ev_op_iterations"],
            ev_op_max_order=tuple(self.config["ev_op_max_order"]),
            sv_mode=self.sv_mode,
            is_threshold=self.is_threshold,
            n3lo_ad_variation=self.config["n3lo_ad_variation"],
            is_polarized=self.config["polarized"],
            is_time_like=self.config["time_like"],
            use_fhmruvv=self.config["use_fhmruvv"],
        )

    def initialize_op_members(self):
        """Init all operators with the identity or zeros."""
        eye = OpMember(
            np.eye(self.grid_size), np.zeros((self.grid_size, self.grid_size))
        )
        zero = OpMember(*[np.zeros((self.grid_size, self.grid_size))] * 2)
        if self.order[1] == 0:
            full_labels = self.full_labels
            non_singlet_labels = br.non_singlet_labels
        else:
            full_labels = self.full_labels_qed
            non_singlet_labels = br.non_singlet_unified_labels
        for n in full_labels:
            if n in self.labels:
                # non-singlet evolution and diagonal op are identities
                if n in non_singlet_labels or n[0] == n[1]:
                    self.op_members[n] = eye.copy()
                else:
                    self.op_members[n] = zero.copy()
            else:
                self.op_members[n] = zero.copy()

    def run_op_integration(
        self,
        log_grid,
    ):
        """Run the integration for each grid point.

        Parameters
        ----------
        log_grid : tuple(k, logx)
            log grid point with relative index

        Returns
        -------
        list
            computed operators at the give grid point
        """
        column = []
        k, logx = log_grid
        start_time = time.perf_counter()
        # iterate basis functions
        for j, bf in enumerate(self.int_disp):
            if k == j and j == self.grid_size - 1:
                continue
            temp_dict = {}
            # iterate sectors
            for label in self.labels:
                res = integrate.quad(
                    self.quad_ker(
                        label=label, logx=logx, areas=bf.areas_representation
                    ),
                    0.5,
                    1.0 - self._mellin_cut,
                    epsabs=1e-12,
                    epsrel=1e-5,
                    limit=100,
                    full_output=1,
                )
                temp_dict[label] = res[:2]
            column.append(temp_dict)
        logger.info(
            "%s: computing operators - %u/%u took: %6f s",
            self.log_label,
            k + 1,
            self.grid_size,
            (time.perf_counter() - start_time),
        )
        return column

    def compute(self):
        """Compute the actual operators (i.e. run the integrations)."""
        self.initialize_op_members()

        # skip computation?
        if np.isclose(self.q2_from, self.q2_to):
            # unless we have to do some scale variation
            # TODO remove if K is factored out of here
            if not (
                self.sv_mode == sv.Modes.expanded
                and not np.isclose(self.xif2, 1.0)
                and not self.is_threshold
            ):
                logger.info(
                    "%s: skipping unity operator at %e", self.log_label, self.q2_from
                )
                self.copy_ns_ops()
                return

        logger.info(
            "%s: computing operators %e -> %e, nf=%d",
            self.log_label,
            self.q2_from,
            self.q2_to,
            self.nf,
        )
        logger.info(
            "%s: µ_R^2 distance: %e -> %e",
            self.log_label,
            self.mu2[0],
            self.mu2[1],
        )
        if self.sv_mode != sv.Modes.unvaried:
            logger.info(
                "Scale Variation: (µ_F/µ_R)^2 = %e, mode: %s",
                self.xif2,
                self.sv_mode.name,
            )
        logger.info(
            "%s: a_s distance: %e -> %e", self.log_label, self.a_s[0], self.a_s[1]
        )
        if self.order[1] > 0:
            logger.info(
                "%s: a_em distance: %e -> %e",
                self.log_label,
                self.a_em[0],
                self.a_em[1],
            )
        logger.info(
            "%s: order: (%d, %d), solution strategy: %s, use fhmruvv: %s",
            self.log_label,
            self.order[0],
            self.order[1],
            self.config["method"],
            self.config["use_fhmruvv"],
        )

        self.integrate()
        # copy non-singlet kernels, if necessary
        self.copy_ns_ops()

    def integrate(
        self,
    ):
        """Run the integration."""
        tot_start_time = time.perf_counter()

        # run integration in parallel for each grid point
        # or avoid opening a single pool
        args = (self.run_op_integration, enumerate(np.log(self.int_disp.xgrid.raw)))
        if self.n_pools == 1:
            res = map(*args)
        else:
            with Pool(self.n_pools) as pool:
                res = pool.map(*args)

        # collect results
        for j, row in enumerate(res):
            for k, entry in enumerate(row):
                for label, (val, err) in entry.items():
                    self.op_members[label].value[j][k] = val
                    self.op_members[label].error[j][k] = err

        # closing comment
        logger.info(
            "%s: Total time %f s",
            self.log_label,
            time.perf_counter() - tot_start_time,
        )

    def copy_ns_ops(self):
        """Copy non-singlet kernels, if necessary."""
        if self.order[1] == 0:
            if self.order[0] == 1:  # in LO +=-=v
                for label in ["nsV", "ns-"]:
                    self.op_members[
                        (br.non_singlet_pids_map[label], 0)
                    ].value = self.op_members[
                        (br.non_singlet_pids_map["ns+"], 0)
                    ].value.copy()
                    self.op_members[
                        (br.non_singlet_pids_map[label], 0)
                    ].error = self.op_members[
                        (br.non_singlet_pids_map["ns+"], 0)
                    ].error.copy()
            elif self.order[0] == 2:  # in NLO -=v
                self.op_members[
                    (br.non_singlet_pids_map["nsV"], 0)
                ].value = self.op_members[
                    (br.non_singlet_pids_map["ns-"], 0)
                ].value.copy()
                self.op_members[
                    (br.non_singlet_pids_map["nsV"], 0)
                ].error = self.op_members[
                    (br.non_singlet_pids_map["ns-"], 0)
                ].error.copy()
        # at O(as0aem1) u-=u+, d-=d+
        # starting from O(as1aem1) P+ != P-
        # However the solution with pure QED is not implemented in EKO
        # so the ns anomalous dimensions are always different
        # elif self.order[1] == 1 and self.order[0] == 0:
        #     self.op_members[
        #         (br.non_singlet_pids_map["ns-u"], 0)
        #     ].value = self.op_members[(br.non_singlet_pids_map["ns+u"], 0)].value.copy()
        #     self.op_members[
        #         (br.non_singlet_pids_map["ns-u"], 0)
        #     ].error = self.op_members[(br.non_singlet_pids_map["ns+u"], 0)].error.copy()
        #     self.op_members[
        #         (br.non_singlet_pids_map["ns-d"], 0)
        #     ].value = self.op_members[(br.non_singlet_pids_map["ns+d"], 0)].value.copy()
        #     self.op_members[
        #         (br.non_singlet_pids_map["ns-d"], 0)
        #     ].error = self.op_members[(br.non_singlet_pids_map["ns+d"], 0)].error.copy()



r"""Integration kernels."""

import logging

import numba as nb
import numpy as np

from .. import interpolation
from .. import scale_variations as sv
from ..io.types import InversionMethod
from ..kernels import non_singlet as ns
from ..kernels import non_singlet_qed as qed_ns
from ..kernels import singlet as s
from ..kernels import singlet_qed as qed_s
from ..kernels import valence_qed as qed_v
from ..matchings import lepton_number
from ..scale_variations import expanded as sv_expanded
from ..scale_variations import exponentiated as sv_exponentiated

logger = logging.getLogger(__name__)


@nb.njit(cache=True)
def select_singlet_element(ker, mode0, mode1):
    """Select element of the singlet matrix.

    Parameters
    ----------
    ker : numpy.ndarray
        singlet integration kernel
    mode0 : int
        id for first sector element
    mode1 : int
        id for second sector element

    Returns
    -------
    complex
        singlet integration kernel element
    """
    j = 0 if mode0 == 100 else 1
    k = 0 if mode1 == 100 else 1
    return ker[j, k]


CB_SIGNATURE = nb.types.double(
    nb.types.CPointer(nb.types.double),  # re_*_raw
    nb.types.CPointer(nb.types.double),  # im_*_raw
    nb.types.double,  # re_n
    nb.types.double,  # im_n
    nb.types.double,  # re_jac
    nb.types.double,  # im_jac
    nb.types.uintc,  # order_qcd
    nb.types.uintc,  # order_qed
    nb.types.bool_,  # is_singlet
    nb.types.uintc,  # mode0
    nb.types.uintc,  # mode1
    nb.types.uintc,  # nf
    nb.types.bool_,  # is_log
    nb.types.double,  # logx
    nb.types.CPointer(nb.types.double),  # areas_raw
    nb.types.uintc,  # areas_x
    nb.types.uintc,  # areas_y
    nb.types.uintc,  # method_num
    nb.types.double,  # as1
    nb.types.double,  # as0
    nb.types.uintc,  # ev_op_iterations
    nb.types.uintc,  # ev_op_max_order_qcd
    nb.types.uintc,  # sv_mode_num
    nb.types.bool_,  # is_threshold
    nb.types.double,  # Lsv
    nb.types.CPointer(nb.types.double),  # as_list
    nb.types.uintc,  # as_list_len
    nb.types.double,  # mu2_from
    nb.types.double,  # mu2_to
    nb.types.CPointer(nb.types.double),  # a_half
    nb.types.uintc,  # a_half_x
    nb.types.uintc,  # a_half_y
    nb.types.bool_,  # alphaem_running
)


@nb.cfunc(
    CB_SIGNATURE,
    cache=True,
    nopython=True,
)
def cb_quad_ker_qcd(
    re_gamma_raw,
    im_gamma_raw,
    re_n,
    im_n,
    re_jac,
    im_jac,
    order_qcd,
    _order_qed,
    is_singlet,
    mode0,
    mode1,
    nf,
    is_log,
    logx,
    areas_raw,
    areas_x,
    areas_y,
    ev_method,
    as1,
    as0,
    ev_op_iterations,
    ev_op_max_order_qcd,
    sv_mode,
    is_threshold,
    Lsv,
    _as_list,
    _as_list_len,
    _mu2_from,
    _mu2_to,
    _a_half,
    _a_half_x,
    _a_half_y,
    _alphaem_running,
):
    """C Callback inside integration kernel."""
    # recover complex variables
    n = re_n + im_n * 1j
    jac = re_jac + im_jac * 1j
    # combute basis functions
    areas = nb.carray(areas_raw, (areas_x, areas_y))
    pj = interpolation.evaluate_grid(n, is_log, logx, areas)
    order = (order_qcd, 0)
    ev_op_max_order = (ev_op_max_order_qcd, 0)
    if is_singlet:
        # reconstruct singlet matrices
        re_gamma_singlet = nb.carray(re_gamma_raw, (order_qcd, 2, 2))
        im_gamma_singlet = nb.carray(im_gamma_raw, (order_qcd, 2, 2))
        gamma_singlet = re_gamma_singlet + im_gamma_singlet * 1j
        if sv_mode == sv.Modes.exponentiated:
            gamma_singlet = sv_exponentiated.gamma_variation(
                gamma_singlet, order, nf, Lsv
            )
        # construct eko
        ker = s.dispatcher(
            order,
            ev_method,
            gamma_singlet,
            as1,
            as0,
            nf,
            ev_op_iterations,
            ev_op_max_order,
        )
        # scale var expanded is applied on the kernel
        if sv_mode == sv.Modes.expanded and not is_threshold:
            ker = np.ascontiguousarray(
                sv_expanded.singlet_variation(gamma_singlet, as1, order, nf, Lsv, dim=2)
            ) @ np.ascontiguousarray(ker)
        ker = select_singlet_element(ker, mode0, mode1)
    else:
        # construct non-singlet matrices
        re_gamma_ns = nb.carray(re_gamma_raw, order_qcd)
        im_gamma_ns = nb.carray(im_gamma_raw, order_qcd)
        gamma_ns = re_gamma_ns + im_gamma_ns * 1j
        if sv_mode == sv.Modes.exponentiated:
            gamma_ns = sv_exponentiated.gamma_variation(gamma_ns, order, nf, Lsv)
        # construct eko
        ker = ns.dispatcher(
            order,
            ev_method,
            gamma_ns,
            as1,
            as0,
            nf,
            ev_op_iterations,
        )
        if sv_mode == sv.Modes.expanded and not is_threshold:
            ker = sv_expanded.non_singlet_variation(gamma_ns, as1, order, nf, Lsv) * ker
    # recombine everything
    res = ker * pj * jac
    return np.real(res)


@nb.njit(cache=True)
def build_ome(A, matching_order, a_s, backward_method):
    r"""Construct the matching expansion in :math:`a_s` with the appropriate
    method.

    Parameters
    ----------
    A : numpy.ndarray
        list of |OME|
    matching_order : tuple(int,int)
        perturbation matching order
    a_s : float
        strong coupling, needed only for the exact inverse
    backward_method : InversionMethod or None
        empty or method for inverting the matching condition (exact or expanded)

    Returns
    -------
    ome : numpy.ndarray
        matching operator matrix
    """
    # to get the inverse one can use this FORM snippet
    # Symbol a;
    # NTensor c,d,e;
    # Local x=-(a*c+a**2* d + a**3 * e);
    # Local bi = 1+x+x**2+x**3;
    # Print;
    # .end
    ome = np.eye(len(A[0]), dtype=np.complex_)
    A = A[:, :, :]
    A = np.ascontiguousarray(A)
    if backward_method is InversionMethod.EXPANDED:
        # expended inverse
        if matching_order[0] >= 1:
            ome -= a_s * A[0]
        if matching_order[0] >= 2:
            ome += a_s**2 * (-A[1] + A[0] @ A[0])
        if matching_order[0] >= 3:
            ome += a_s**3 * (-A[2] + A[0] @ A[1] + A[1] @ A[0] - A[0] @ A[0] @ A[0])
    else:
        # forward or exact inverse
        if matching_order[0] >= 1:
            ome += a_s * A[0]
        if matching_order[0] >= 2:
            ome += a_s**2 * A[1]
        if matching_order[0] >= 3:
            ome += a_s**3 * A[2]
        # need inverse exact ?  so add the missing pieces
        if backward_method is InversionMethod.EXACT:
            ome = np.linalg.inv(ome)
    return ome


@nb.cfunc(
    CB_SIGNATURE,
    cache=True,
    nopython=True,
)
def cb_quad_ker_ome(
    re_ome_raw,
    im_ome_raw,
    re_n,
    im_n,
    re_jac,
    im_jac,
    order_qcd,
    _order_qed,
    is_singlet,
    mode0,
    mode1,
    nf,
    is_log,
    logx,
    areas_raw,
    areas_x,
    areas_y,
    backward_method,
    as1,
    _as0,
    _ev_op_iterations,
    _ev_op_max_order_qcd,
    sv_mode,
    _is_threshold,
    Lsv,
    _as_list,
    _as_list_len,
    _mu2_from,
    _mu2_to,
    _a_half,
    _a_half_x,
    _a_half_y,
    _alphaem_running,
):
    """C Callback inside integration kernel."""
    # recover complex variables
    n = re_n + im_n * 1j
    jac = re_jac + im_jac * 1j
    # compute basis functions
    areas = nb.carray(areas_raw, (areas_x, areas_y))
    pj = interpolation.evaluate_grid(n, is_log, logx, areas)
    order = (order_qcd, 0)
    if is_singlet:
        indices = {21: 0, 100: 1, 90: 2}
        # reconstruct singlet matrices
        re_ome_singlet = nb.carray(re_ome_raw, (order_qcd, 3, 3))
        im_ome_singlet = nb.carray(im_ome_raw, (order_qcd, 3, 3))
        A = re_ome_singlet + im_ome_singlet * 1j
    else:
        indices = {200: 0, 91: 1}
        # construct non-singlet matrices
        re_ome_ns = nb.carray(re_ome_raw, (order_qcd, 2, 2))
        im_ome_ns = nb.carray(im_ome_raw, (order_qcd, 2, 2))
        A = re_ome_ns + im_ome_ns * 1j

    # correct for scale variations
    if sv_mode == sv.Modes.exponentiated:
        A = sv.exponentiated.gamma_variation(A, order, nf, Lsv)

    # build the expansion in alpha_s depending on the strategy
    ker = build_ome(A, order, as1, backward_method)

    # select the needed matrix element
    ker = ker[indices[mode0], indices[mode1]]

    # recombine everything
    res = ker * pj * jac
    return np.real(res)


@nb.njit(cache=True)
def select_QEDsinglet_element(ker, mode0, mode1):
    """Select element of the QEDsinglet matrix.

    Parameters
    ----------
    ker : numpy.ndarray
        QEDsinglet integration kernel
    mode0 : int
        id for first sector element
    mode1 : int
        id for second sector element
    Returns
    -------
    ker : complex
        QEDsinglet integration kernel element
    """
    if mode0 == 21:
        index1 = 0
    elif mode0 == 22:
        index1 = 1
    elif mode0 == 100:
        index1 = 2
    else:
        index1 = 3
    if mode1 == 21:
        index2 = 0
    elif mode1 == 22:
        index2 = 1
    elif mode1 == 100:
        index2 = 2
    else:
        index2 = 3
    return ker[index1, index2]


@nb.njit(cache=True)
def select_QEDvalence_element(ker, mode0, mode1):
    """Select element of the QEDvalence matrix.

    Parameters
    ----------
    ker : numpy.ndarray
        QEDvalence integration kernel
    mode0 : int
        id for first sector element
    mode1 : int
        id for second sector element
    Returns
    -------
    ker : complex
        QEDvalence integration kernel element
    """
    index1 = 0 if mode0 == 10200 else 1
    index2 = 0 if mode1 == 10200 else 1
    return ker[index1, index2]


@nb.cfunc(
    CB_SIGNATURE,
    cache=True,
    nopython=True,
)
def cb_quad_ker_qed(
    re_gamma_raw,
    im_gamma_raw,
    re_n,
    im_n,
    re_jac,
    im_jac,
    order_qcd,
    order_qed,
    is_singlet,
    mode0,
    mode1,
    nf,
    is_log,
    logx,
    areas_raw,
    areas_x,
    areas_y,
    ev_method,
    _as1,
    _as0,
    ev_op_iterations,
    ev_op_max_order_qcd,
    sv_mode,
    is_threshold,
    Lsv,
    as_list_raw,
    as_list_len,
    mu2_from,
    mu2_to,
    a_half_raw,
    a_half_x,
    a_half_y,
    alphaem_running,
):
    """C Callback inside integration kernel."""
    # recover complex variables
    n = re_n + im_n * 1j
    jac = re_jac + im_jac * 1j
    # compute basis functions
    areas = nb.carray(areas_raw, (areas_x, areas_y))
    pj = interpolation.evaluate_grid(n, is_log, logx, areas)
    order = (order_qcd, order_qed)
    ev_op_max_order = (ev_op_max_order_qcd, order_qed)
    is_valence = (mode0 == 10200) or (mode0 == 10204)

    as_list = nb.carray(as_list_raw, as_list_len)
    a_half = nb.carray(a_half_raw, (a_half_x, a_half_y))

    if is_singlet:
        # reconstruct singlet matrices
        re_gamma_singlet = nb.carray(re_gamma_raw, (order_qcd + 1, order_qed + 1, 4, 4))
        im_gamma_singlet = nb.carray(im_gamma_raw, (order_qcd + 1, order_qed + 1, 4, 4))
        gamma_singlet = re_gamma_singlet + im_gamma_singlet * 1j

        # scale var exponentiated is directly applied on gamma
        if sv_mode == sv.Modes.exponentiated:
            gamma_singlet = sv.exponentiated.gamma_variation_qed(
                gamma_singlet, order, nf, lepton_number(mu2_to), Lsv, alphaem_running
            )

        ker = qed_s.dispatcher(
            order,
            ev_method,
            gamma_singlet,
            as_list,
            a_half,
            nf,
            ev_op_iterations,
            ev_op_max_order,
        )
        if sv_mode == sv.Modes.expanded and not is_threshold:
            ker = np.ascontiguousarray(
                sv.expanded.singlet_variation_qed(
                    gamma_singlet,
                    as_list[-1],
                    a_half[-1][1],
                    alphaem_running,
                    order,
                    nf,
                    Lsv,
                )
            ) @ np.ascontiguousarray(ker)
        ker = select_QEDsinglet_element(ker, mode0, mode1)

    elif is_valence:
        # reconstruct valence matrices
        re_gamma_valence = nb.carray(re_gamma_raw, (order_qcd + 1, order_qed + 1, 2, 2))
        im_gamma_valence = nb.carray(im_gamma_raw, (order_qcd + 1, order_qed + 1, 2, 2))
        gamma_valence = re_gamma_valence + im_gamma_valence * 1j

        if sv_mode == sv.Modes.exponentiated:
            gamma_valence = sv.exponentiated.gamma_variation_qed(
                gamma_valence, order, nf, lepton_number(mu2_to), Lsv, alphaem_running
            )
        ker = qed_v.dispatcher(
            order,
            ev_method,
            gamma_valence,
            as_list,
            a_half,
            nf,
            ev_op_iterations,
            ev_op_max_order,
        )
        # scale var expanded is applied on the kernel
        if sv_mode == sv.Modes.expanded and not is_threshold:
            ker = np.ascontiguousarray(
                sv.expanded.valence_variation_qed(
                    gamma_valence,
                    as_list[-1],
                    a_half[-1][1],
                    alphaem_running,
                    order,
                    nf,
                    Lsv,
                )
            ) @ np.ascontiguousarray(ker)
        ker = select_QEDvalence_element(ker, mode0, mode1)

    else:
        # construct non-singlet matrices
        re_gamma_ns = nb.carray(re_gamma_raw, (order_qcd + 1, order_qed + 1))
        im_gamma_ns = nb.carray(im_gamma_raw, (order_qcd + 1, order_qed + 1))
        gamma_ns = re_gamma_ns + im_gamma_ns * 1j
        if sv_mode == sv.Modes.exponentiated:
            gamma_ns = sv_exponentiated.gamma_variation_qed(
                gamma_ns, order, nf, lepton_number(mu2_to), Lsv, alphaem_running
            )
        # construct eko
        ker = qed_ns.dispatcher(
            order,
            ev_method,
            gamma_ns,
            as_list,
            a_half[:, 1],
            alphaem_running,
            nf,
            ev_op_iterations,
            mu2_from,
            mu2_to,
        )
        if sv_mode == sv.Modes.expanded and not is_threshold:
            ker = (
                sv_expanded.non_singlet_variation_qed(
                    gamma_ns,
                    as_list[-1],
                    a_half[-1][1],
                    alphaem_running,
                    order,
                    nf,
                    Lsv,
                )
                * ker
            )
    # recombine everything
    res = ker * pj * jac
    return np.real(res)



"""Fully managed runner.

This is an automated runner, mainly suggested for small EKOs
computations.

The primitives used here to compute the various pieces are part of the
public interface, and should be directly used to manage a more complex
run for a considebaly large operator.

Thus, parallelization and multi-node execution is possible using EKO
primitives, but not automatically performed.
"""

from pathlib import Path

from ..io.items import Target
from ..io.runcards import OperatorCard, TheoryCard
from ..io.struct import EKO
from . import operators, parts, recipes


def solve(theory: TheoryCard, operator: OperatorCard, path: Path):
    """Solve DGLAP equations in terms of evolution kernel operators (EKO)."""
    with EKO.create(path) as builder:
        eko = builder.load_cards(theory, operator).build()  # pylint: disable=E1101
        recipes.create(eko)

        for recipe in eko.recipes:
            eko.parts[recipe] = parts.evolve(eko, recipe)
            # flush the memory
            del eko.parts[recipe]
        for recipe in eko.recipes_matching:
            eko.parts_matching[recipe] = parts.match(eko, recipe)
            # flush the memory
            del eko.parts_matching[recipe]

        for ep in operator.evolgrid:
            components = operators.retrieve(ep, eko)
            target = Target.from_ep(ep)
            eko.operators[target] = operators.join(components)
            # flush the memory
            del eko.parts
            del eko.operators[target]



"""Physical and internal objects.

This subpackage contains the definitions of many objects, with the
purpose of using them as type hints, but not only (at least for some of
them).

Few related functions are also defined here.
"""



"""The solutions to the |DGLAP| equations."""

import enum

from ..io.types import EvolutionMethod


class EvoMethods(enum.IntEnum):
    """Enumerate evolution methods."""

    ITERATE_EXACT = enum.auto()
    ITERATE_EXPANDED = enum.auto()
    PERTURBATIVE_EXACT = enum.auto()
    PERTURBATIVE_EXPANDED = enum.auto()
    TRUNCATED = enum.auto()
    ORDERED_TRUNCATED = enum.auto()
    DECOMPOSE_EXACT = enum.auto()
    DECOMPOSE_EXPANDED = enum.auto()


def ev_method(s: EvolutionMethod) -> EvoMethods:
    """Return the evolution method.

    Parameters
    ----------
    s :
        string representation

    Returns
    -------
    i :
        int representation
    """
    return EvoMethods[s.value.upper().replace("-", "_")]



"""The |N3LO| Altarelli-Parisi splitting kernels.

For further documentation see :doc:`N3LO anomalous dimensions
<../../../theory/N3LO_ad>`
"""

import numba as nb
import numpy as np

from . import fhmruvv as fhmruvv
from .ggg import gamma_gg
from .ggq import gamma_gq
from .gnsm import gamma_nsm
from .gnsp import gamma_nsp
from .gnsv import gamma_nsv
from .gps import gamma_ps
from .gqg import gamma_qg


@nb.njit(cache=True)
def gamma_singlet(N, nf, cache, variation):
    r"""Compute the |N3LO| singlet anomalous dimension matrix.

      .. math::
          \gamma_S^{(3)} = \left(\begin{array}{cc}
          \gamma_{qq}^{(3)} & \gamma_{qg}^{(3)}\\
          \gamma_{gq}^{(3)} & \gamma_{gg}^{(3)}
          \end{array}\right)

    Parameters
    ----------
    N : complex
      Mellin moment
    nf : int
      Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache
    variation : tuple
        |N3LO| anomalous dimension variation ``(gg, gq, qg, qq)``

    Returns
    -------
    numpy.ndarray
        |N3LO| singlet anomalous dimension matrix
        :math:`\gamma_{S}^{(3)}(N)`
    """
    gamma_qq = gamma_nsp(N, nf, cache) + gamma_ps(N, nf, cache, variation[3])
    gamma_S_0 = np.array(
        [
            [gamma_qq, gamma_qg(N, nf, cache, variation[2])],
            [
                gamma_gq(N, nf, cache, variation[1]),
                gamma_gg(N, nf, cache, variation[0]),
            ],
        ],
        np.complex_,
    )
    return gamma_S_0


@nb.njit(cache=True)
def gamma_singlet_qed(N, nf, cache, variation):
    r"""Compute the leading-order singlet anomalous dimension matrix for the
    unified evolution basis.

    .. math::
        \\gamma_S^{(3,0)} = \\left(\begin{array}{cccc}
        \\gamma_{gg}^{(3,0)} & 0 & \\gamma_{gq}^{(3,0)} & 0\\
        0 & 0 & 0 & 0 \\
        \\gamma_{qg}^{(3,0)} & 0 & \\gamma_{qq}^{(3,0)} & 0 \\
        0 & 0 & 0 & \\gamma_{qq}^{(3,0)} \\
        \\end{array}\right)

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    numpy.ndarray
        Leading-order singlet anomalous dimension matrix :math:`\\gamma_{S}^{(3,0)}(N)`
    """
    gamma_np_p = gamma_nsp(N, nf, cache)
    gamma_qq = gamma_np_p + gamma_ps(N, nf, cache, variation[3])
    gamma_S = np.array(
        [
            [
                gamma_gg(N, nf, cache, variation[0]),
                0.0 + 0.0j,
                gamma_gq(N, nf, cache, variation[1]),
                0.0 + 0.0j,
            ],
            [0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j],
            [gamma_qg(N, nf, cache, variation[2]), 0.0 + 0.0j, gamma_qq, 0.0 + 0.0j],
            [0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j, gamma_np_p],
        ],
        np.complex_,
    )
    return gamma_S


@nb.njit(cache=True)
def gamma_valence_qed(N, nf, cache):
    r"""Compute the leading-order valence anomalous dimension matrix for the
    unified evolution basis.

    .. math::
        \\gamma_V^{(3,0)} = \\left(\begin{array}{cc}
        \\gamma_{nsV}^{(3,0)} & 0\\
        0 & \\gamma_{ns-}^{(3,0)}
        \\end{array}\right)

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    numpy.ndarray
        Leading-order singlet anomalous dimension matrix :math:`\\gamma_{V}^{(3,0)}(N)`
    """
    gamma_V = np.array(
        [
            [gamma_nsv(N, nf, cache), 0.0],
            [0.0, gamma_nsm(N, nf, cache)],
        ],
        np.complex_,
    )
    return gamma_V



"""Types and quantities related to theory couplings."""

import dataclasses
import enum

from ..io.dictlike import DictLike
from ..io.types import EvolutionPoint as EPoint
from ..io.types import ReferenceRunning, Scalar

Coupling = Scalar
CouplingRef = ReferenceRunning[Coupling]


@dataclasses.dataclass
class CouplingsInfo(DictLike):
    """Reference values for coupling constants."""

    alphas: Coupling
    r"""Reference |QCD| coupling :math:`\alpha_s(\mu_R^0)^{(n_f^0)}`.

    Note
    ----
    We refer to :math:`\alpha_s` here instead
    of :math:`a_s = \alpha_s/(4\pi)` as we do otherwise.
    """
    alphaem: Coupling
    r"""Reference |QED| coupling :math:`\alpha_{em}(\mu_R^0)^{(n_f^0)}`."""
    ref: EPoint
    r"""Reference evolution point :math:`(\mu_R^0,n_f^0)`"""
    em_running: bool = False
    r"""If set, activates the |QED| running."""

    @property
    def values(self):
        """Collect only couplings values."""
        return (self.alphas, self.alphaem)


# TODO: upgrade the following to StrEnum (requires py>=3.11) with that, it is
# possible to replace all non-alias right sides with calls to enum.auto()


class CouplingEvolutionMethod(enum.Enum):
    """Beta functions solution method."""

    EXACT = "exact"
    EXPANDED = "expanded"



"""Compute operator components.

.. todo::

    This is the only part of the new runner making use of the global context
    through the :class:`EKO` object.

    After #242, the goal is to update :class:`Operator` and
    :class:`OperatorMatrixElement` to simplify the computation and passing down
    parameters.
"""

import numpy as np

from .. import evolution_operator as evop
from ..evolution_operator import matching_condition, physical
from ..evolution_operator import operator_matrix_element as ome
from ..io import EKO
from ..io.items import Evolution, Matching, Operator
from ..quantities.heavy_quarks import QuarkMassScheme
from . import commons


def _managers(eko: EKO) -> evop.Managers:
    """Collect managers for operator computation.

    .. todo::

        Legacy interface, avoid managers usage.
    """
    tcard = eko.theory_card
    ocard = eko.operator_card
    return evop.Managers(
        atlas=commons.atlas(tcard, ocard),
        couplings=commons.couplings(tcard, ocard),
        interpolator=commons.interpolator(ocard),
    )


def _evolve_configs(eko: EKO) -> dict:
    """Create configs for :class:`Operator`.

    .. todo::

        Legacy interface, make use of a dedicated object.
    """
    tcard = eko.theory_card
    ocard = eko.operator_card
    return dict(
        order=tcard.order,
        xif2=tcard.xif**2,
        method=ocard.configs.evolution_method.value,
        ev_op_iterations=ocard.configs.ev_op_iterations,
        ev_op_max_order=ocard.configs.ev_op_max_order,
        polarized=ocard.configs.polarized,
        time_like=ocard.configs.time_like,
        debug_skip_singlet=ocard.debug.skip_singlet,
        debug_skip_non_singlet=ocard.debug.skip_non_singlet,
        n_integration_cores=ocard.configs.n_integration_cores,
        ModSV=ocard.configs.scvar_method,
        n3lo_ad_variation=tcard.n3lo_ad_variation,
        use_fhmruvv=tcard.use_fhmruvv,
        # Here order is shifted by one, no QED matching is available so far.
        matching_order=tcard.matching_order,
    )


def evolve(eko: EKO, recipe: Evolution) -> Operator:
    """Compute evolution in isolation."""
    op = evop.Operator(
        _evolve_configs(eko),
        _managers(eko),
        recipe.as_atlas,
        is_threshold=recipe.cliff,
    )
    op.compute()

    qed = eko.theory_card.order[1] > 0
    res, err = physical.PhysicalOperator.ad_to_evol_map(
        op.op_members, op.nf, op.q2_to, qed
    ).to_flavor_basis_tensor(qed)

    return Operator(res, err)


def _matching_configs(eko: EKO) -> dict:
    """Create configs for :class:`OperatorMatrixElement`.

    .. todo::

        Legacy interface, make use of a dedicated object.
    """
    ocard = eko.operator_card
    return dict(
        **_evolve_configs(eko),
        backward_inversion=ocard.configs.inversion_method,
    )


def match(eko: EKO, recipe: Matching) -> Operator:
    """Compute matching in isolation.

    Note
    ----
    Compared to the old prescription, a dedicated rotation to a different
    intrinsic basis is not needed any longer.

    All the operators are blown up to flavor basis, and they are saved and
    joined in that unique basis. So, the only rotation used is towards that
    basis, and encoded in the blowing up prescription.
    """
    kthr = eko.theory_card.heavy.squared_ratios[recipe.hq - 4]
    op = ome.OperatorMatrixElement(
        _matching_configs(eko),
        _managers(eko),
        recipe.hq - 1,
        recipe.scale,
        recipe.inverse,
        np.log(kthr),
        eko.theory_card.heavy.masses_scheme is QuarkMassScheme.MSBAR,
    )
    op.compute()
    qed = eko.theory_card.order[1] > 0
    res, err = matching_condition.MatchingCondition.split_ad_to_evol_map(
        op.op_members, op.nf, recipe.scale, qed
    ).to_flavor_basis_tensor(qed)

    return Operator(res, err)



"""Structures to hold runcards information.

All energy scales in the runcards should be saved linearly, not the
squared value, for consistency. Squares are consistently taken inside.
"""

from dataclasses import dataclass
from math import nan
from typing import List, Optional, Union

import numpy as np
import numpy.typing as npt

from .. import basis_rotation as br
from .. import interpolation, msbar_masses
from .. import version as vmod
from ..couplings import couplings_mod_ev
from ..matchings import Atlas, nf_default
from ..quantities import heavy_quarks as hq
from ..quantities.couplings import CouplingsInfo
from ..quantities.heavy_quarks import HeavyInfo, QuarkMassScheme
from .dictlike import DictLike
from .types import (
    EvolutionMethod,
    InversionMethod,
    N3LOAdVariation,
    Order,
    RawCard,
    ScaleVariationsMethod,
    SquaredScale,
    T,
)
from .types import EvolutionPoint as EPoint


# TODO: add frozen
@dataclass
class TheoryCard(DictLike):
    """Represents general theory settings.

    This describes the general settings, which define the parameters for the general
    framework rather than a specific operator.
    Important settings are e.g. perturbative order, quark masses, or
    the reference value of the couplings.
    """

    order: Order
    r"""Perturbative order tuple, ``(QCD, QED)``.

    The entries correspond to the power of the respective couplings,
    e.g. ``(1,0)`` refers to :math:`a_s^1 a_{em}^0`,
    i.e. |LO| |QCD| evolution.
    """
    couplings: CouplingsInfo
    """Couplings configuration."""
    heavy: HeavyInfo
    """Heavy quarks related information."""
    xif: float
    """Ratio between factorization scale and process scale."""
    n3lo_ad_variation: N3LOAdVariation
    """|N3LO| anomalous dimension variation: ``(gg, gq, qg, qq, nsp, nsm,
    nsv)``."""
    use_fhmruvv: Optional[bool] = True
    """If True use the |FHMRUVV| |N3LO| anomalous dimensions."""
    matching_order: Optional[Order] = None
    """Matching conditions perturbative order tuple, ``(QCD, QED)``.

    Similar, to order the entries refer to the powers of the respective couplings.
    |QED| |OME| are currently not available.
    Note that for |LO| |QCD| evolution one needs ``(0,0)`` as ``matching_order``,
    because the matching conditions are shifted by one order
    (since they are related to processes).
    If not provided it will use this assumption as default.
    """

    def __post_init__(self):
        """Enforce defaults."""
        if self.matching_order is None:
            self.matching_order = (self.order[0] - 1, 0)


@dataclass
class Debug(DictLike):
    """Debug configurations."""

    skip_singlet: bool = False
    """Whether to skip singlet computation."""
    skip_non_singlet: bool = False
    """Whether to skip non-singlet computation."""


@dataclass
class Configs(DictLike):
    """Solution specific configurations."""

    evolution_method: EvolutionMethod
    """Evolution mode."""
    ev_op_max_order: Order
    """Maximum order to use in U matrices expansion.

    Used only in ``perturbative`` solutions.
    """
    ev_op_iterations: int
    """Number of intervals in which to break the evolution integral."""
    scvar_method: Optional[ScaleVariationsMethod]
    """Scale variation method."""
    inversion_method: Optional[InversionMethod]
    """Which method to use for backward matching conditions."""
    interpolation_polynomial_degree: int
    """Degree of elements of the interpolation polynomial basis."""
    interpolation_is_log: bool
    r"""Whether to use polynomials in :math:`\log(x)`.

    If `false`, polynomials are in :math:`x`.
    """
    polarized: bool
    """If `true` do polarized evolution."""
    time_like: bool
    """If `true` do time-like evolution."""
    n_integration_cores: int = 1
    """Number of cores used to parallelize integration."""


@dataclass
class OperatorCard(DictLike):
    """Represents specific operator settings.

    This describes the specific settings for a given set of operator,
    rather than the general theory settings.
    Important settings are e.g. the initial and final evolution points,
    or the solution method.
    """

    init: EPoint
    """Initial evolution point."""
    mugrid: List[EPoint]
    """Final list of evolution points."""
    xgrid: interpolation.XGrid
    """Momentum fraction interpolation grid."""

    # collections
    configs: Configs
    """Specific configuration to be used during the calculation of these
    operators."""
    debug: Debug
    """Debug configurations."""

    # optional
    eko_version: str = vmod.__version__
    """Version of EKO package used for generation."""

    # a few properties, for ease of use and compatibility
    @property
    def mu20(self):
        """Squared value of initial scale."""
        return self.init[0] ** 2

    @property
    def mu2grid(self) -> npt.NDArray:
        """Grid of squared final scales."""
        return np.array([mu for mu, _ in self.mugrid]) ** 2

    @property
    def evolgrid(self) -> List[EPoint]:
        """Grid of squared final evolution points."""
        return [(mu**2, nf) for mu, nf in self.mugrid]

    @property
    def pids(self):
        """Internal flavor basis, used for computation."""
        return np.array(br.flavor_basis_pids)


Card = Union[TheoryCard, OperatorCard]


@dataclass
class Legacy:
    """Upgrade legacy runcards."""

    theory: RawCard
    operator: RawCard

    MOD_EV2METHOD = {
        "EXA": "iterate-exact",
        "EXP": "iterate-expanded",
        "TRN": "truncated",
    }

    @staticmethod
    def heavies(pattern: str, old_th: dict):
        """Retrieve a set of values for all heavy flavors."""
        return [old_th[pattern % q] for q in hq.FLAVORS]

    @staticmethod
    def fallback(*args: T, default: Optional[T] = None) -> T:
        """Return the first not None argument."""
        for arg in args:
            if arg is not None:
                return arg

        if default is None:
            return args[-1]
        return default

    @property
    def new_theory(self):
        """Build new format theory runcard."""
        old = self.theory
        new = {}

        new["order"] = [old["PTO"] + 1, old["QED"]]
        alphaem = self.fallback(old.get("alphaqed"), old.get("alphaem"), default=0.0)
        if "Qedref" in old:
            em_running = bool(np.isclose(old["Qedref"], old["Qref"]))
        else:
            em_running = False
        ms = self.heavies("m%s", old)
        new["couplings"] = dict(
            alphas=old["alphas"],
            alphaem=alphaem,
            em_running=em_running,
            ref=(old["Qref"], old["nfref"]),
        )
        new["heavy"] = {
            "matching_ratios": self.heavies("k%sThr", old),
            "masses_scheme": old["HQ"],
        }
        if old["HQ"] == "POLE":
            new["heavy"]["masses"] = [[m, nan] for m in ms]
        elif old["HQ"] == "MSBAR":
            mus = self.heavies("Qm%s", old)
            new["heavy"]["masses"] = [[m, mu] for m, mu in zip(ms, mus)]
        else:
            raise ValueError(f"Unknown mass scheme '{old['HQ']}'")

        new["xif"] = old["XIF"]

        new["n3lo_ad_variation"] = old.get("n3lo_ad_variation", (0, 0, 0, 0, 0, 0, 0))
        # here PTO: 0 means truly LO, no QED matching is available so far.
        new["matching_order"] = old.get("PTO_matching", [old["PTO"], 0])
        new["use_fhmruvv"] = old.get("use_fhmruvv", True)
        return TheoryCard.from_dict(new)

    @property
    def new_operator(self):
        """Build new format operator runcard."""
        old = self.operator
        old_th = self.theory
        new = {}

        ms = self.heavies("m%s", old_th)
        ks = self.heavies("k%sThr", old_th)
        new["init"] = (
            old_th["Q0"],
            (
                nf_default(old_th["Q0"] ** 2.0, default_atlas(ms, ks))
                if old_th["nf0"] is None
                else old_th["nf0"]
            ),
        )
        if "mugrid" in old:
            mugrid = old["mugrid"]
        else:
            mu2grid = old["Q2grid"] if "Q2grid" in old else old["mu2grid"]
            mugrid = np.sqrt(mu2grid).tolist()
        new["mugrid"] = flavored_mugrid(mugrid, list(ms), list(ks))

        new["configs"] = {}
        evmod = old_th["ModEv"]
        new["configs"]["evolution_method"] = self.MOD_EV2METHOD.get(evmod, evmod)
        new["configs"]["inversion_method"] = old_th.get(
            "backward_inversion", "expanded"
        )
        new["configs"]["scvar_method"] = old_th.get("ModSV", "expanded")
        for k in (
            "interpolation_polynomial_degree",
            "interpolation_is_log",
            "ev_op_iterations",
            "n_integration_cores",
            "polarized",
            "time_like",
        ):
            new["configs"][k] = old[k]
        max_order = old["ev_op_max_order"]
        if isinstance(max_order, int):
            new["configs"]["ev_op_max_order"] = [max_order, old_th["QED"]]

        new["debug"] = {}
        lpref = len("debug_")
        for k in ("debug_skip_non_singlet", "debug_skip_singlet"):
            new["debug"][k[lpref:]] = old[k]

        new["xgrid"] = old["interpolation_xgrid"]

        return OperatorCard.from_dict(new)


def default_atlas(masses: list, matching_ratios: list):
    r"""Create default threshold landscape.

    This method should not be used to write new runcards, but rather to
    have a consistent default for comparison with other softwares and
    existing |PDF| sets. There is no one-to-one relation between number of
    running flavors and final scales, unless matchings are all applied.
    But this is a custom choice, since it is possible to have |PDF| in
    different |FNS| at the same scales.
    """
    matchings = (np.array(masses) * np.array(matching_ratios)) ** 2
    return Atlas(matchings.tolist(), (0.0, 0))


def flavored_mugrid(mugrid: list, masses: list, matching_ratios: list):
    r"""Upgrade :math:`\mu^2` grid to contain also target number flavors.

    It determines the number of flavors for the |PDF| set at the target
    scale, inferring it according to the specified scales.

    This method should not be used to write new runcards, but rather to
    have a consistent default for comparison with other softwares and
    existing |PDF| sets. There is no one-to-one relation between number of
    running flavors and final scales, unless matchings are all applied.
    But this is a custom choice, since it is possible to have |PDF| in
    different |FNS| at the same scales.
    """
    atlas = default_atlas(masses, matching_ratios)
    return [(mu, nf_default(mu**2, atlas)) for mu in mugrid]


# TODO: move to a more suitable place
def masses(theory: TheoryCard, evmeth: EvolutionMethod) -> List[SquaredScale]:
    """Compute masses in the chosen scheme."""
    if theory.heavy.masses_scheme is QuarkMassScheme.MSBAR:
        return msbar_masses.compute(
            theory.heavy.masses,
            theory.couplings,
            theory.order,
            couplings_mod_ev(evmeth),
            np.power(theory.heavy.matching_ratios, 2.0).tolist(),
            xif2=theory.xif**2,
        ).tolist()
    if theory.heavy.masses_scheme is QuarkMassScheme.POLE:
        return [mq.value**2 for mq in theory.heavy.masses]

    raise ValueError(f"Unknown mass scheme '{theory.heavy.masses_scheme}'")



"""Launch EKO calculations."""

import pathlib
from typing import Sequence

import click
import yaml

import eko
from eko.io.runcards import OperatorCard, TheoryCard

from .base import command
from .library import OPERATOR, OUTPUT, THEORY


@command.command("run")
@click.argument("paths", nargs=-1, type=click.Path(path_type=pathlib.Path))
def subcommand(paths: Sequence[pathlib.Path]):
    r"""Launch calculation specified by PATHS.

    Compute the operator given runcards. How they are specified depends on the
    amount of arguments:

        1. only one argument provided: it is used as the folder where runcards
        and output are located (with default names)

        2. two arguments: they are used as theory and operator card paths, and
        the output is placed in the same folder of the operator card

        3. three arguments: same as two, but the third argument is used as
        output path


    Default names are:

    - theory card: "theory.yaml"
    - operator card: "operator.yaml"
    - output: "eko.tar".
    """
    if len(paths) == 1:
        theory = paths[0] / THEORY
        operator = paths[0] / OPERATOR
    elif len(paths) in [2, 3]:
        theory = paths[0]
        operator = paths[1]
    else:
        raise click.UsageError(
            f"Only one to three arguments allowed, '{len(paths)}' passed."
        )

    if len(paths) == 3:
        output = paths[2]
    else:
        output = operator.parent / OUTPUT

    tc = yaml.safe_load(theory.read_text(encoding="utf-8"))
    if "order" in tc:
        tc = TheoryCard.from_dict(tc)
    oc = yaml.safe_load(operator.read_text(encoding="utf-8"))
    if "configs" in oc:
        oc = OperatorCard.from_dict(oc)

    eko.solve(tc, oc, path=output)



r"""Compute the beta function coefficients.

The strong coupling :math:`a_s(\mu^2)` and the electro-magnetic coupling
:math:`a_{em}(\mu^2)` obey separate, but coupled |RGE|:

.. math::

    \mu^2 \frac{da_s(\mu^2)}{d\mu^2} = \beta_{QCD}^{(n_f)}(a_s(\mu^2), a_{em}(\mu^2))
        = - \sum_{j=2}\sum_{k=0} \beta_{QCD}^{(j,k),(n_f)}
                  \left(a_s(\mu^2)\right)^j \left(a_{em}(\mu^2)\right)^k\\

    \mu^2 \frac{da_{em}(\mu^2)}{d\mu^2} = \beta_{QED}^{(n_f)}(a_s(\mu^2), a_{em}(\mu^2))
        = - \sum_{j=0}\sum_{k=2} \beta_{QED}^{(j,k),(n_f)}
                  \left(a_s(\mu^2)\right)^j \left(a_{em}(\mu^2)\right)^k

See the :mod:`eko.couplings` module for solutions.
When considering QED corrections the two |RGE| need two be solved
simultaneously.
"""

from typing import Tuple

import numba as nb

from . import constants
from .constants import zeta3


@nb.njit(cache=True)
def beta_qcd_as2(nf: int) -> float:
    r"""Compute the first coefficient of the QCD beta function.

    Implements :eqref:`3.1` of :cite:`Herzog:2017ohr`.

    Parameters
    ----------
    nf :
        number of active flavors

    Returns
    -------
    float
        first coefficient of the QCD beta function :math:`\beta_{QCD}^{(2,0),(n_f)}`
    """
    return 11.0 / 3.0 * constants.CA - 4.0 / 3.0 * constants.TR * nf


@nb.njit(cache=True)
def beta_qed_aem2(nf: int, nl: int) -> float:
    r"""Compute the first coefficient of the QED beta function.

    Implements :eqref:`7` of :cite:`Surguladze:1996hx`.

    Parameters
    ----------
    nf :
        number of active flavors
    nl :
        number of leptons

    Returns
    -------
    float
        first coefficient of the QED beta function :math:`\beta_{QED}^{(0,2),(n_f)}`
    """
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    return -4.0 / 3 * (nl + constants.NC * (nu * constants.eu2 + nd * constants.ed2))


@nb.njit(cache=True)
def beta_qcd_as3(nf: int) -> float:
    r"""Compute the second coefficient of the QCD beta function.

    Implements :eqref:`3.2` of :cite:`Herzog:2017ohr`.

    Parameters
    ----------
    nf :
        number of active flavors

    Returns
    -------
    float
        second coefficient of the QCD beta function :math:`\beta_{QCD}^{(3,0),(n_f)}`
    """
    TF = constants.TR * nf
    b_ca2 = 34.0 / 3.0 * constants.CA * constants.CA
    b_ca = -20.0 / 3.0 * constants.CA * TF
    b_cf = -4.0 * constants.CF * TF
    return b_ca2 + b_ca + b_cf


@nb.njit(cache=True)
def beta_qed_aem3(nf: int, nl: int) -> float:
    r"""Compute the second coefficient of the QED beta function.

    Implements :eqref:`7` of :cite:`Surguladze:1996hx`.

    Parameters
    ----------
    nf :
        number of active flavors
    nl :
        number of leptons

    Returns
    -------
    float
        second coefficient of the QED beta function :math:`\beta_{QED}^{(0,3),(n_f)}`
    """
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    return -4.0 * (nl + constants.NC * (nu * constants.eu2**2 + nd * constants.ed2**2))


@nb.njit(cache=True)
def beta_qcd_as2aem1(nf: int) -> float:
    r"""Compute the first QED correction of the QCD beta function.

    Implements :eqref:`7` of :cite:`Surguladze:1996hx`.

    Parameters
    ----------
    nf :
        number of active flavors

    Returns
    -------
    float
        first QED correction of the QCD beta function :math:`\beta_{QCD}^{(2,1),(n_f)}`
    """
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    return -4.0 * constants.TR * (nu * constants.eu2 + nd * constants.ed2)


@nb.njit(cache=True)
def beta_qed_aem2as1(nf: int) -> float:
    r"""Compute the first QCD correction of the QED beta function.

    Implements :eqref:`7` of :cite:`Surguladze:1996hx`.

    Parameters
    ----------
    nf :
        number of active flavors

    Returns
    -------
    float
        first QCD correction of the QED beta function :math:`\beta_{QED}^{(1,2),(n_f)}`
    """
    nu = constants.uplike_flavors(nf)
    nd = nf - nu
    return (
        -4.0 * constants.CF * constants.NC * (nu * constants.eu2 + nd * constants.ed2)
    )


@nb.njit(cache=True)
def beta_qcd_as4(nf: int) -> float:
    r"""Compute the third coefficient of the QCD beta function.

    Implements :eqref:`3.3` of :cite:`Herzog:2017ohr`.

    Parameters
    ----------
    nf :
        number of active flavors

    Returns
    -------
    float
        third coefficient of the QCD beta function :math:`\beta_{QCD}^{(4,0),(n_f)}`
    """
    TF = constants.TR * nf
    return (
        2857.0 / 54.0 * constants.CA * constants.CA * constants.CA
        - 1415.0 / 27.0 * constants.CA * constants.CA * TF
        - 205.0 / 9.0 * constants.CF * constants.CA * TF
        + 2.0 * constants.CF * constants.CF * TF
        + 44.0 / 9.0 * constants.CF * TF * TF
        + 158.0 / 27.0 * constants.CA * TF * TF
    )


@nb.njit(cache=True)
def beta_qcd_as5(nf: int) -> float:
    r"""Compute the fourth coefficient of the QCD beta function.

    Implements :eqref:`3.6` of :cite:`Herzog:2017ohr`.

    Parameters
    ----------
    nf :
        number of active flavors

    Returns
    -------
    float
        fourth coefficient of the QCD beta function :math:`\beta_{QCD}^{(5,0),(n_f)}`
    """
    return (
        149753.0 / 6.0
        + 3564.0 * zeta3
        + nf * (-1078361.0 / 162.0 - 6508.0 / 27.0 * zeta3)
        + nf**2 * (50065.0 / 162.0 + 6472.0 / 81.0 * zeta3)
        + 1093.0 / 729.0 * nf**3
    )


@nb.njit(cache=True)
def beta_qcd(k: Tuple[int, int], nf: int) -> float:
    r"""Compute QCD beta coefficient.

    Parameters
    ----------
    k :
        perturbative orders
    nf :
        number of active flavors

    Returns
    -------
    float
        :math:`\beta_{QCD}^{k,(nf)}`
    """
    beta_ = 0
    if k == (2, 0):
        beta_ = beta_qcd_as2(nf)
    elif k == (3, 0):
        beta_ = beta_qcd_as3(nf)
    elif k == (4, 0):
        beta_ = beta_qcd_as4(nf)
    elif k == (5, 0):
        beta_ = beta_qcd_as5(nf)
    elif k == (2, 1):
        beta_ = beta_qcd_as2aem1(nf)
    else:
        raise ValueError("Beta_QCD coefficients beyond N3LO are not implemented!")
    return beta_


@nb.njit(cache=True)
def beta_qed(k: Tuple[int, int], nf: int, nl: int) -> float:
    r"""Compute QED beta coefficient.

    Parameters
    ----------
    k :
        perturbative order
    nf :
        number of active flavors
    nl :
        number of leptons

    Returns
    -------
    float
        :math:`\beta_{QED}^{k,(nf)}`
    """
    beta_ = 0
    if k == (0, 2):
        beta_ = beta_qed_aem2(nf, nl)
    elif k == (0, 3):
        beta_ = beta_qed_aem3(nf, nl)
    elif k == (1, 2):
        beta_ = beta_qed_aem2as1(nf)
    else:
        raise ValueError("Beta_QED coefficients beyond NLO are not implemented!")
    return beta_


@nb.njit(cache=True)
def b_qcd(k: Tuple[int, int], nf: int) -> float:
    r"""Compute normalized QCD beta coefficient.

    Parameters
    ----------
    k :
        perturbative order
    nf :
        number of active flavors

    Returns
    -------
    float
        :math:`b_{QCD}^{k,(nf)} = \beta_{QCD}^{k,(nf)} / \beta_{QCD}^{(2,0),(nf)}`
    """
    return beta_qcd(k, nf) / beta_qcd((2, 0), nf)


@nb.njit(cache=True)
def b_qed(k: Tuple[int, int], nf: int, nl: int) -> float:
    r"""Compute normalized QED beta coefficient.

    Parameters
    ----------
    k :
        perturbative order
    nf :
        number of active flavors
    nl :
        number of leptons

    Returns
    -------
    float
        :math:`b_{QED}^{k,(nf)} = \beta_{QED}^{k,(nf)} / \beta_{QED}^{(0,2),(nf)}`
    """
    return beta_qed(k, nf, nl) / beta_qed((0, 2), nf, nl)



"""Apply evolution operator to a PDF."""

from collections.abc import Sequence
from typing import Optional

import numpy as np
import numpy.typing as npt

from eko import basis_rotation as br
from eko import interpolation
from eko.io import EKO
from eko.io.types import EvolutionPoint

RawPdfResult = dict[EvolutionPoint, npt.ArrayLike]
"""PDFs as raw grids.

The key is given by the associated evolution point. The values are
tensors sorted by (replica, flavor, xgrid). It may be the PDF or the
associated integration error.
"""


LabeledPdfResult = dict[EvolutionPoint, dict[int, npt.ArrayLike]]
"""PDFs labeled by their PDF identifier.

The outer key is given by the associated evolution point. The inner key
is the |PID|. The inner values are the values for along the xgrid. It
may be the PDF or the associated integration error.
"""


def apply_pdf(
    eko: EKO,
    lhapdf_like,
    targetgrid: npt.ArrayLike = None,
    rotate_to_evolution_basis: bool = False,
) -> tuple[LabeledPdfResult, LabeledPdfResult]:
    """Apply all available operators to the input PDF.

    Parameters
    ----------
    eko :
        eko output object containing all informations
    lhapdf_like : Any
        object that provides an `xfxQ2` callable (as `lhapdf <https://lhapdf.hepforge.org/>`_
        and :class:`ekomark.toyLH.toyPDF` do) (and thus is in flavor basis)
    targetgrid :
        if given, interpolates to the targetgrid (instead of xgrid)
    rotate_to_evolution_basis :
        if True rotate to evoluton basis

    Returns
    -------
    pdfs :
        PDFs for the computed evolution points
    errors :
        Integration errors for PDFs for the computed evolution points
    """
    # prepare post-process
    qed = eko.theory_card.order[1] > 0
    flavor_rotation = None
    labels = br.flavor_basis_pids
    if rotate_to_evolution_basis:
        if not qed:
            flavor_rotation = br.rotate_flavor_to_evolution
            labels = br.evol_basis_pids
        else:
            flavor_rotation = br.rotate_flavor_to_unified_evolution
            labels = br.unified_evol_basis_pids
    return apply_pdf_flavor(eko, lhapdf_like, labels, targetgrid, flavor_rotation)


def apply_pdf_flavor(
    eko: EKO,
    lhapdf_like,
    flavor_labels: Sequence[int],
    targetgrid: npt.ArrayLike = None,
    flavor_rotation: npt.ArrayLike = None,
) -> tuple[LabeledPdfResult, LabeledPdfResult]:
    """Apply all available operators to the input PDF.

    Parameters
    ----------
    eko :
        eko output object containing all informations
    lhapdf_like : Any
        object that provides an `xfxQ2` callable (as `lhapdf <https://lhapdf.hepforge.org/>`_
        and :class:`ekomark.toyLH.toyPDF` do) (and thus is in flavor basis)
    flavor_labels :
        flavor names
    targetgrid :
        if given, interpolates to the targetgrid (instead of xgrid)
    flavor_rotation :
        if give, rotate in flavor space

    Returns
    -------
    pdfs :
        PDFs for the computed evolution points
    errors :
        Integration errors for PDFs for the computed evolution points
    """
    # create pdfs
    input_pdfs = np.zeros((len(br.flavor_basis_pids), len(eko.xgrid)))
    for j, pid in enumerate(br.flavor_basis_pids):
        if not lhapdf_like.hasFlavor(pid):
            continue
        input_pdfs[j] = np.array(
            [lhapdf_like.xfxQ2(pid, x, eko.mu20) / x for x in eko.xgrid.raw]
        )
    # apply
    grids, grid_errors = apply_grids(eko, input_pdfs[None, :])
    new_grids = rotate_result(eko, grids, flavor_labels, targetgrid, flavor_rotation)
    new_errors = rotate_result(
        eko, grid_errors, flavor_labels, targetgrid, flavor_rotation
    )
    # unwrap the replica axis again
    pdfs: LabeledPdfResult = {}
    errors: LabeledPdfResult = {}
    for ep, pdf in new_grids.items():
        pdfs[ep] = {lab: grid[0] for lab, grid in pdf.items()}
        if ep in new_errors:
            errors[ep] = {lab: (grid[0]) for lab, grid in new_errors[ep].items()}
    return pdfs, errors


def rotate_result(
    eko: EKO,
    grids: RawPdfResult,
    flavor_labels: Sequence[int],
    targetgrid: Optional[npt.ArrayLike] = None,
    flavor_rotation: Optional[npt.ArrayLike] = None,
) -> LabeledPdfResult:
    """Rotate and relabel PDFs.

    Parameters
    ----------
    eko :
        eko output object containing all informations
    grids :
        Raw grids coming from evolution
    flavor_labels :
        flavors names
    targetgrid :
        if given, interpolates to the targetgrid (instead of xgrid)
    flavor_rotation :
        if given, rotates in flavor space

    Returns
    -------
    pdfs :
        relabeled and, if requested rotated, PDFs
    """
    # rotate to evolution basis
    if flavor_rotation is not None:
        new_grids = {}
        for ep, pdf_grid in grids.items():
            new_grids[ep] = np.einsum(
                "ab,rbk->rak", flavor_rotation, pdf_grid, optimize="optimal"
            )
        grids = new_grids

    # rotate/interpolate to target grid
    if targetgrid is not None:
        b = interpolation.InterpolatorDispatcher(
            xgrid=eko.xgrid,
            polynomial_degree=eko.operator_card.configs.interpolation_polynomial_degree,
            mode_N=False,
        )

        x_rotation = b.get_interpolation(targetgrid)
        new_grids = {}
        for ep, pdf_grid in grids.items():
            new_grids[ep] = np.einsum(
                "jk,rbk->rbj", x_rotation, pdf_grid, optimize="optimal"
            )
        grids = new_grids

    # relabel
    new_grids = {}
    for ep, pdf_grid in grids.items():
        new_grids[ep] = dict(
            zip(
                flavor_labels,
                np.swapaxes(pdf_grid, 0, 1),
            )
        )
    grids = new_grids

    return grids


_EKO_CONTRACTION = "ajbk,rbk->raj"
"""Contract eko for all replicas."""


def apply_grids(
    eko: EKO, input_grids: npt.ArrayLike
) -> tuple[RawPdfResult, RawPdfResult]:
    """Apply all available operators to the input grids.

    Parameters
    ----------
    eko :
        eko output object
    input_grids :
        3D PDF grids evaluated at the inital scale. The axis have to be (replica, flavor, xgrid)

    Returns
    -------
    pdfs :
        output PDFs for the computed evolution points
    errors :
        associated integration errors for the computed evolution points
    """
    # sanity check
    if len(input_grids.shape) != 3 or input_grids.shape[1:] != (
        len(br.flavor_basis_pids),
        len(eko.xgrid),
    ):
        raise ValueError(
            "input grids have to be sorted by replica, flavor, xgrid!"
            f"The shape has to be (r,{len(br.flavor_basis_pids)},{len(eko.xgrid)})"
        )
    # iterate
    pdfs: RawPdfResult = {}
    errors: RawPdfResult = {}
    for ep, elem in eko.items():
        pdfs[ep] = np.einsum(
            _EKO_CONTRACTION, elem.operator, input_grids, optimize="optimal"
        )
        if elem.error is not None:
            errors[ep] = np.einsum(
                _EKO_CONTRACTION, elem.error, input_grids, optimize="optimal"
            )
    return pdfs, errors



"""Defines the matching conditions for the |VFNS| evolution."""

from .. import basis_rotation as br
from .. import member


class MatchingCondition(member.OperatorBase):
    """Matching conditions for |PDF| at threshold.

    The continuation of the (formally) heavy non-singlet distributions
    with either the full singlet :math:`S` or the full valence :math:`V`
    is considered "trivial". Instead at |NNLO| additional terms ("non-trivial")
    enter.
    """

    @classmethod
    def split_ad_to_evol_map(
        cls,
        op_members,
        nf,
        q2_thr,
        qed=False,
    ):
        """Create the instance from the |OME|.

        Parameters
        ----------
            op_members : eko.operator_matrix_element.OperatorMatrixElement.op_members
                Attribute of :class:`~eko.operator_matrix_element.OperatorMatrixElement`
                containing the |OME|
            nf : int
                number of active flavors *below* the threshold
            q2_thr: float
                threshold value
            qed : bool
                activate qed
        """
        m = {
            "S.S": op_members[(100, 100)],
            "S.g": op_members[(100, 21)],
            "g.S": op_members[(21, 100)],
            "g.g": op_members[(21, 21)],
            "V.V": op_members[(200, 200)],
        }

        # add elements which are already active
        if not qed:
            for f in range(2, nf + 1):
                n = f**2 - 1
                m[f"V{n}.V{n}"] = m["V.V"]
                m[f"T{n}.T{n}"] = m["V.V"]
        else:
            m.update(
                {
                    "Sdelta.Sdelta": op_members[(200, 200)],
                    "Vdelta.Vdelta": op_members[(200, 200)],
                    "ph.ph": member.OpMember.id_like(op_members[(200, 200)]).copy(),
                }
            )
            names = {3: "d3", 4: "u3", 5: "d8", 6: "u8"}
            for k in range(3, nf + 1):
                m[f"V{names[k]}.V{names[k]}"] = m["V.V"]
                m[f"T{names[k]}.T{names[k]}"] = m["V.V"]

        # activate the next heavy quark
        hq = br.quark_names[nf]
        m.update(
            {
                # f"{hq}-.V": op_members[(br.matching_hminus_pid, 200)],
                f"{hq}+.S": op_members[(br.matching_hplus_pid, 100)],
                f"{hq}+.g": op_members[(br.matching_hplus_pid, 21)],
            }
        )

        # intrinsic matching
        op_id = member.OpMember.id_like(op_members[(200, 200)])
        for intr_fl in [4, 5, 6]:
            ihq = br.quark_names[intr_fl - 1]  # find name
            if intr_fl > nf + 1:
                # keep the higher quarks as they are
                m[f"{ihq}+.{ihq}+"] = op_id.copy()
                m[f"{ihq}-.{ihq}-"] = op_id.copy()
            elif intr_fl == nf + 1:
                # match the missing contribution from h+ and h-
                m.update(
                    {
                        f"{ihq}+.{ihq}+": op_members[
                            (br.matching_hplus_pid, br.matching_hplus_pid)
                        ],
                        f"S.{ihq}+": op_members[(100, br.matching_hplus_pid)],
                        f"g.{ihq}+": op_members[(21, br.matching_hplus_pid)],
                        f"{ihq}-.{ihq}-": op_members[
                            (br.matching_hminus_pid, br.matching_hminus_pid)
                        ],
                        # f"V.{ihq}-": op_members[(200, br.matching_hminus_pid)],
                    }
                )
        return cls.promote_names(m, q2_thr)



"""Manipulate output generate by EKO."""

import copy
import logging
import warnings
from typing import Callable, Optional

import numpy as np
import numpy.typing as npt

from .. import basis_rotation as br
from .. import interpolation
from ..interpolation import XGrid
from .struct import Operator

logger = logging.getLogger(__name__)

TARGETGRID_ROTATION = "ij,ajbk->aibk"
INPUTGRID_ROTATION = "ajbk,kl->ajbl"
SIMGRID_ROTATION = "ij,ajbk,kl->aibl"
"""Simultaneous grid rotation contraction indices."""


def rotation(
    new: Optional[XGrid], old: XGrid, check: Callable, compute: Callable
) -> npt.NDArray:
    """Define grid rotation.

    This function returns the necessary rotation,
    if the checks for a non-trivial new grid are passed.

    However, the check and the computation are delegated respectively to the
    callables `check` and `compute`.
    """
    if new is None:
        return None

    if check(new, old):
        warnings.warn("The new grid is close to the current one")
        return None

    return compute(new, old)


def xgrid_check(new: Optional[XGrid], old: XGrid) -> bool:
    """Check validity of new xgrid."""
    return new is not None and len(new) == len(old) and np.allclose(new.raw, old.raw)


def xgrid_compute_rotation(
    new: XGrid, old: XGrid, interpdeg: int, swap: bool = False
) -> npt.NDArray:
    """Compute rotation from old to new xgrid.

    By default, the roation is computed for a target xgrid. Whether the function
    should be used for an input xgrid, the `swap` argument should be set to
    `True`, in order to compute it in the other direction (i.e. the transposed).
    """
    if swap:
        new, old = old, new
    b = interpolation.InterpolatorDispatcher(old, interpdeg, False)
    return b.get_interpolation(new.raw)


def xgrid_reshape(
    elem: Operator,
    xgrid: XGrid,
    interpdeg: int,
    targetgrid: Optional[XGrid] = None,
    inputgrid: Optional[XGrid] = None,
) -> Operator:
    """Reinterpolate the operator on output and/or input grid(s).

    Target corresponds to the output PDF.
    """
    # calling with no arguments is an error
    if targetgrid is None and inputgrid is None:
        raise ValueError("Nor inputgrid nor targetgrid was given")

    check = xgrid_check
    crot = xgrid_compute_rotation

    # construct matrices
    targetrot = rotation(
        targetgrid,
        xgrid,
        check,
        lambda new, old: crot(new, old, interpdeg),
    )
    inputrot = rotation(
        inputgrid,
        xgrid,
        check,
        lambda new, old: crot(new, old, interpdeg, swap=True),
    )
    # after the checks: if there is still nothing to do, skip
    if targetrot is None and inputrot is None:
        logger.debug("Nothing done.")
        return copy.deepcopy(elem)

    # build new grid
    operands = [elem.operator]
    operands_errs = [elem.error]

    if targetrot is not None and inputrot is None:
        contraction = TARGETGRID_ROTATION
    elif inputrot is not None and targetrot is None:
        contraction = INPUTGRID_ROTATION
    else:
        contraction = SIMGRID_ROTATION

    if targetrot is not None:
        operands.insert(0, targetrot)
        operands_errs.insert(0, targetrot)
    if inputrot is not None:
        operands.append(inputrot)
        operands_errs.append(inputrot)

    new_operator = np.einsum(contraction, *operands, optimize="optimal")
    if elem.error is not None:
        new_error = np.einsum(contraction, *operands_errs, optimize="optimal")
    else:
        new_error = None

    return Operator(operator=new_operator, error=new_error)


TARGETPIDS_ROTATION = "ca,ajbk->cjbk"
INPUTPIDS_ROTATION = "ajbk,bd->ajdk"
SIMPIDS_ROTATION = "ca,ajbk,bd->cjdk"
"""Simultaneous grid rotation contraction indices."""


def flavor_reshape(
    elem: Operator,
    targetpids: Optional[npt.NDArray] = None,
    inputpids: Optional[npt.NDArray] = None,
) -> Operator:
    """Change the operator to have in the output targetpids and/or in the input
    inputpids.

    Parameters
    ----------
    elem :
        the operator to be rotated
    targetpids :
        target rotation specified in the flavor basis
    inputpids :
        input rotation specified in the flavor basis
    """
    # calling with no arguments is an error
    if targetpids is None and inputpids is None:
        raise ValueError("Nor inputpids nor targetpids was given")
    # now check to the current status
    if targetpids is not None and np.allclose(
        targetpids, np.eye(elem.operator.shape[0])
    ):
        targetpids = None
        warnings.warn("The new targetpids is close to current basis")
    if inputpids is not None and np.allclose(inputpids, np.eye(elem.operator.shape[2])):
        inputpids = None
        warnings.warn("The new inputpids is close to current basis")
    # after the checks: if there is still nothing to do, skip
    if targetpids is None and inputpids is None:
        logger.debug("Nothing done.")
        return copy.deepcopy(elem)

    # flip input around
    inv_inputpids = np.zeros_like(inputpids)
    if inputpids is not None:
        inv_inputpids = np.linalg.inv(inputpids)

    # build new grid
    ops = elem.operator
    errs = elem.error
    if targetpids is not None and inputpids is None:
        ops = np.einsum(TARGETPIDS_ROTATION, targetpids, ops, optimize="optimal")
        errs = (
            np.einsum(TARGETPIDS_ROTATION, targetpids, errs, optimize="optimal")
            if errs is not None
            else None
        )
    elif inputpids is not None and targetpids is None:
        ops = np.einsum(INPUTPIDS_ROTATION, ops, inv_inputpids, optimize="optimal")
        errs = (
            np.einsum(INPUTPIDS_ROTATION, errs, inv_inputpids, optimize="optimal")
            if errs is not None
            else None
        )
    else:
        ops = np.einsum(
            SIMPIDS_ROTATION, targetpids, ops, inv_inputpids, optimize="optimal"
        )
        errs = (
            np.einsum(
                SIMPIDS_ROTATION,
                targetpids,
                errs,
                inv_inputpids,
                optimize="optimal",
            )
            if errs is not None
            else None
        )

    return Operator(operator=ops, error=errs)


def to_evol(elem: Operator, source: bool = True, target: bool = False) -> Operator:
    """Rotate the operator into evolution basis.

    Parameters
    ----------
    elem :
        the operator to be rotated
    source :
        rotate on the input tensor
    target :
        rotate on the output tensor
    """
    # rotate
    inputpids = br.rotate_flavor_to_evolution if source else None
    targetpids = br.rotate_flavor_to_evolution if target else None
    return flavor_reshape(elem, inputpids=inputpids, targetpids=targetpids)


def to_uni_evol(elem: Operator, source: bool = True, target: bool = False) -> Operator:
    """Rotate the operator into evolution basis.

    Parameters
    ----------
    elem :
        the operator to be rotated
    source :
        rotate on the input tensor
    target :
        rotate on the output tensor
    """
    # rotate
    inputpids = br.rotate_flavor_to_unified_evolution if source else None
    targetpids = br.rotate_flavor_to_unified_evolution if target else None
    return flavor_reshape(elem, inputpids=inputpids, targetpids=targetpids)



r"""Holds the classes that define the |FNS|."""

import logging
from dataclasses import dataclass
from typing import List, Union

import numba as nb
import numpy as np

from .constants import MTAU
from .io.types import EvolutionPoint as EPoint
from .io.types import FlavorIndex, FlavorsNumber, SquaredScale
from .quantities.heavy_quarks import MatchingScales

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Segment:
    """Oriented path in the threshold landscape."""

    origin: SquaredScale
    """Starting point."""
    target: SquaredScale
    """Final point."""
    nf: FlavorsNumber
    """Number of active flavors."""

    @property
    def is_downward(self) -> bool:
        """Return True if ``origin`` is bigger than ``target``."""
        return self.origin > self.target

    def __str__(self):
        """Textual representation, mainly for logging purpose."""
        return f"Segment({self.origin} -> {self.target}, nf={self.nf})"


Path = List[Segment]


@dataclass(frozen=True)
class Matching:
    """Matching between two different segments.

    The meaning of the flavor index `hq` is the PID of the corresponding heavy
    quark.
    """

    scale: SquaredScale
    hq: FlavorIndex
    inverse: bool


MatchedPath = List[Union[Segment, Matching]]


class Atlas:
    r"""Holds information about the matching scales.

    These scales are the :math:`Q^2` has to pass in order to get there
    from a given :math:`Q^2_{ref}`.
    """

    def __init__(self, matching_scales: MatchingScales, origin: EPoint):
        """Create basic atlas."""
        self.walls = [0] + matching_scales + [np.inf]
        self.origin = self.normalize(origin)

        logger.info(str(self))

    def normalize(self, target: EPoint) -> EPoint:
        """Fill number of flavors if needed."""
        if target[1] is not None:
            return target
        return (target[0], nf_default(target[0], self))

    def __str__(self):
        """Textual representation, mainly for logging purpose."""
        walls = " - ".join([f"{w:.2e}" for w in self.walls])
        return f"Atlas [{walls}], ref={self.origin[0]} @ {self.origin[1]}"

    @classmethod
    def ffns(cls, nf: int, mu2: SquaredScale):
        """Create a |FFNS| setup.

        The function creates simply sufficient thresholds at ``0`` (in the
        beginning), since the number of flavors is determined by counting
        from below.

        The origin is set with that number of flavors.
        """
        matching_scales = MatchingScales([0] * (nf - 3) + [np.inf] * (6 - nf))
        origin = (mu2, nf)
        return cls(matching_scales, origin)

    def path(self, target: EPoint) -> Path:
        """Determine the path to the target evolution point.

        Essentially, the path is always monotonic in the number of flavors,
        increasing or decreasing the active flavors by one unit every time a
        matching happens at the suitable scale.

        Examples
        --------
        Since this can result in a counter-intuitive behavior, let's walk through some examples.

        Starting with the intuitive one:
        >>> Atlas([10, 20, 30], (5, 3)).path((25, 5))
        [Segment(5, 10, 3), Segment(10, 20, 4), Segment(20, 25, 5)]

        If the number of flavor has been reached, it will continue walking
        without matchin again.
        >>> Atlas([10, 20, 30], (5, 3)).path((25, 4))
        [Segment(5, 10, 3), Segment(10, 25, 4)]

        It is irrelevant the scale you start from, to step from 3 to 4 you have
        to cross the charm matching scale, whether this means walking upward or
        downward.
        >>> Atlas([10, 20, 30], (15, 3)).path((25, 5))
        [Segment(15, 10, 3), Segment(10, 20, 4), Segment(20, 25, 5)]

        An actual backward evolution is defined by lowering the number of
        flavors going from origin to target.
        >>> Atlas([10, 20, 30], (25, 5)).path((5, 3))
        [Segment(25, 20, 5), Segment(20, 10, 4), Segment(10, 5, 3)]

        But the only difference is in the matching between two segments, since
        a single segment is always happening in a fixed number of flavors, and
        it is completely analogue for upward or downward evolution.

        Note
        ----

        Since the only task required to determine a path is interleaving the
        correct matching scales, this is done by slicing the walls.
        """
        mu20, nf0 = self.origin
        mu2f, nff = self.normalize(target)

        # determine direction and python slice modifier
        rc, shift = (-1, -3) if nff < nf0 else (1, -2)

        # join all necessary points in one list
        boundaries = [mu20] + self.walls[nf0 + shift : nff + shift : rc] + [mu2f]

        return [
            Segment(boundaries[i], mu2, nf0 + i * rc)
            for i, mu2 in enumerate(boundaries[1:])
        ]

    def matched_path(self, target: EPoint) -> MatchedPath:
        """Determine the path to the target, including matchings.

        In practice, just a wrapper around :meth:`path` adding the intermediate
        matchings.
        """
        path = self.path(target)
        inverse = is_downward_path(path)

        prev = path[0]
        matched: MatchedPath = [prev]
        for seg in path[1:]:
            matching = Matching(prev.target, max(prev.nf, seg.nf), inverse)

            matched.append(matching)
            matched.append(seg)
            prev = seg

        return matched


def nf_default(mu2: SquaredScale, atlas: Atlas) -> FlavorsNumber:
    r"""Determine the number of active flavors in the *default flow*.

    Default flow is defined by the natural sorting of the matching scales:

    .. math::

        \mu_c < \mu_b < \mu_t

    So, the flow is defined starting with 3 flavors below the charm matching,
    and increasing by one every time a matching scale is passed while
    increasing the scale.
    """
    ref_idx = np.digitize(mu2, atlas.walls)
    return int(2 + ref_idx)


def is_downward_path(path: Path) -> bool:
    r"""Determine if a path is downward.

    Criterias are:

    - in the number of active flavors when the path list contains more than one
      :class:`Segment`, note this can be different from each
      :attr:`Segment.is_downward`
    - in :math:`\mu^2`, when just one single :class:`Segment` is given
    """
    if len(path) == 1:
        return path[0].is_downward
    return path[1].nf < path[0].nf


def flavor_shift(is_downward: bool) -> int:
    """Determine the shift to number of light flavors."""
    return 4 if is_downward else 3


@nb.njit(cache=True)
def lepton_number(q2):
    """Compute the number of leptons.

    Note: muons and electrons are always massless as for up, down and strange.

    Parameters
    ----------
    q2 : float
        scale

    Returns
    -------
    int :
       Number of leptons
    """
    return 3 if q2 > MTAU**2 else 2



"""Heavy quarks related quantities."""

import enum
from dataclasses import dataclass
from typing import Generic, List, Sequence, TypeVar

import numpy as np

from ..io.dictlike import DictLike
from ..io.types import LinearScale, ReferenceRunning, SquaredScale

FLAVORS = "cbt"

T = TypeVar("T")


class HeavyQuarks(list, Generic[T]):
    """Access heavy quarks attributes by name."""

    def __init__(self, args: Sequence[T]):
        if len(args) != 3:
            raise ValueError("Pass values for exactly three quarks.")

        self.extend(args)

    @property
    def c(self) -> T:
        """Charm quark."""
        return self[0]

    @c.setter
    def c(self, value: T):
        self[0] = value

    @property
    def b(self) -> T:
        """Bottom quark."""
        return self[1]

    @b.setter
    def b(self, value: T):
        self[1] = value

    @property
    def t(self) -> T:
        """Top quark."""
        return self[2]

    @t.setter
    def t(self, value: T):
        self[2] = value


QuarkMass = LinearScale
QuarkMassRef = ReferenceRunning[QuarkMass]
HeavyQuarkMasses = HeavyQuarks[QuarkMassRef]
MatchingRatio = float
MatchingRatios = HeavyQuarks[MatchingRatio]
MatchingScale = SquaredScale
MatchingScales = HeavyQuarks[MatchingScale]


# TODO: upgrade the following to StrEnum (requires py>=3.11) with that, it is
# possible to replace all non-alias right sides with calls to enum.auto()


class QuarkMassScheme(enum.Enum):
    """Scheme to define heavy quark masses."""

    MSBAR = "msbar"
    POLE = "pole"


@dataclass
class HeavyInfo(DictLike):
    """Collect information about heavy quarks.

    This is meant to be used mainly as a theory card section, and to be passed
    around when all or a large part of this information is required.

    Note
    ----
    All scales and ratios in this structure are linear, so you can consider
    them as quantities in :math:`GeV` or ratios of them.
    """

    masses: HeavyQuarkMasses
    """List of heavy quark masses."""
    masses_scheme: QuarkMassScheme
    """Scheme used to specify heavy quark masses."""
    matching_ratios: MatchingRatios
    """Matching scale of heavy quark masses."""

    @property
    def squared_ratios(self) -> List[float]:
        """Squared ratios of matching scales."""
        return np.power(self.matching_ratios, 2.0).tolist()



"""LHAPDF info file utilities."""

import copy
import math
from typing import Any, Dict

import numpy as np

from eko import couplings
from eko.io.runcards import OperatorCard, TheoryCard

from .genpdf import load
from .utils import regroup_evolgrid


def build(
    theory_card: TheoryCard,
    operators_card: OperatorCard,
    num_members: int,
    info_update: dict,
) -> dict:
    """Generate a lhapdf info file.

    Parameters
    ----------
    theory_card :
        theory card
    operators_card :
        operators card
    num_members :
        number of pdf set members
    info_update :
        additional info to update

    Returns
    -------
    dict
        info file in lhapdf format
    """
    template_info = copy.deepcopy(load.template_info)
    template_info["SetDesc"] = (
        "Evolved PDF from " + str(operators_card.init[0]) + " GeV"
    )
    template_info["Authors"] = ""
    template_info["FlavorScheme"] = "variable"
    template_info.update(info_update)
    template_info["NumFlavors"] = max(nf for _, nf in operators_card.mugrid)
    template_info["Flavors"] = [-6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 21, 22]
    # TODO actually point to input grid
    template_info["XMin"] = float(operators_card.xgrid.raw[0])
    template_info["XMax"] = float(operators_card.xgrid.raw[-1])
    template_info["NumMembers"] = num_members
    template_info["OrderQCD"] = theory_card.order[0] - 1
    template_info["QMin"] = round(math.sqrt(operators_card.mu2grid[0]), 4)
    template_info["QMax"] = round(math.sqrt(operators_card.mu2grid[-1]), 4)
    template_info["MZ"] = theory_card.couplings.ref[0]
    template_info["MUp"] = 0.0
    template_info["MDown"] = 0.0
    template_info["MStrange"] = 0.0
    template_info["MCharm"] = theory_card.heavy.masses.c.value
    template_info["MBottom"] = theory_card.heavy.masses.b.value
    template_info["MTop"] = theory_card.heavy.masses.t.value
    # dump alphas
    template_info.update(build_alphas(theory_card, operators_card))
    return template_info


def build_alphas(
    theory_card: TheoryCard,
    operators_card: OperatorCard,
) -> dict:
    """Generate a couplings section of lhapdf info file.

    Parameters
    ----------
    theory_card : dict
        theory card
    operators_card : dict
        operators card

    Returns
    -------
    dict
        info file section in lhapdf format
    """
    # start with meta stuff
    template_info: Dict[str, Any] = {}
    template_info["AlphaS_MZ"] = theory_card.couplings.alphas
    template_info["AlphaS_OrderQCD"] = theory_card.order[0] - 1
    # prepare
    evolgrid = regroup_evolgrid(operators_card.mugrid)
    evmod = couplings.couplings_mod_ev(operators_card.configs.evolution_method)
    quark_masses = [(x.value) ** 2 for x in theory_card.heavy.masses]
    sc = couplings.Couplings(
        theory_card.couplings,
        theory_card.order,
        evmod,
        quark_masses,
        hqm_scheme=theory_card.heavy.masses_scheme,
        thresholds_ratios=np.power(list(iter(theory_card.heavy.matching_ratios)), 2.0),
    )
    # add actual values
    alphas_values = []
    alphas_qs = []
    for nf, mus in evolgrid.items():
        for mu in mus:
            alphas_values.append(float(4.0 * np.pi * sc.a_s(mu * mu, nf_to=nf)))
            alphas_qs.append(mu)

    template_info["AlphaS_Vals"] = alphas_values
    template_info["AlphaS_Qs"] = alphas_qs
    return template_info



"""Tools to evolve actual PDFs."""

import pathlib

import numpy as np

from eko import basis_rotation as br
from eko.io import EKO
from eko.runner import managed

from . import apply, genpdf, info_file
from .utils import regroup_evolgrid

DEFAULT_NAME = "eko.tar"


def evolve_pdfs(
    initial_PDF_list,
    theory_card,
    operators_card,
    path=None,
    store_path=None,
    targetgrid=None,
    install=False,
    name="Evolved_PDF",
    info_update=None,
):
    """Evolves one or several initial_PDFs and dump the evolved PDFs in lhapdf
    format.

    Parameters
    ----------
    initial_PDF_list : list(lhapdf object)
        list of PDF members to be evolved
    theory_card : dict
        theory card
    operators_card : dict
        operators card
    path : str
        path to cached eko output (if "None" it will be recomputed)
    store_path : str
        path where the eko is stored (if "None" will not be saved)
    targetgrid : list(float)
        target x-grid (if different from input x-grid)
    install : bool
        set whether to install evolved PDF to lhapdf directory
    name : str
        set name of evolved PDF
    info_update : dict
        dict of info to add or update to default info file
    """
    # separate by nf the evolgrid (and order per nf/q)
    q2block_per_nf = regroup_evolgrid(operators_card.mugrid)

    # check we have disjoint scale ranges
    nfs = list(q2block_per_nf.keys())
    for j in range(len(nfs) - 1):
        # equal points are allowed by LHAPDF
        if q2block_per_nf[nfs[j]][-1] > q2block_per_nf[nfs[j + 1]][0]:
            raise ValueError(
                f"Last scale point for nf={nfs[j]} is bigger than first in nf={nfs[j + 1]}"
            )

    # update op and th cards
    if path is not None:
        eko_path = pathlib.Path(path)
        if eko_path.is_dir():
            eko_path = eko_path / DEFAULT_NAME
    else:
        if store_path is None:
            raise ValueError("'store_path' required if 'path' is not provided.")
        managed.solve(theory_card, operators_card, path=store_path)
        eko_path = store_path

    # apply PDF to eko
    evolved_PDF_list = []
    q2block_per_nf = {}
    with EKO.read(eko_path) as eko_output:
        for initial_PDF in initial_PDF_list:
            evolved_PDF_list.append(
                apply.apply_pdf(eko_output, initial_PDF, targetgrid)[0]
            )

        # separate by nf the evolgrid (and order per nf/q)
        q2block_per_nf = regroup_evolgrid(eko_output.evolgrid)

    # update info file
    if targetgrid is None:
        targetgrid = operators_card.xgrid
    if info_update is None:
        info_update = {}
    info_update["XMin"] = targetgrid.raw[0]
    info_update["XMax"] = targetgrid.raw[-1]
    info = info_file.build(
        theory_card,
        operators_card,
        len(evolved_PDF_list),
        info_update=info_update,
    )

    # write all replicas
    all_member_blocks = []
    for evolved_PDF in evolved_PDF_list:
        all_blocks = collect_blocks(
            evolved_PDF, q2block_per_nf, targetgrid.raw.tolist()
        )
        all_member_blocks.append(all_blocks)
    genpdf.export.dump_set(name, info, all_member_blocks)

    if install:
        genpdf.install_pdf(name)


def collect_blocks(evolved_PDF: dict, q2block_per_nf: dict, xgrid: list):
    """Collect all LHAPDF blocks for a given replica.

    Parameters
    ----------
    evolved_PDF :
        PDF evaluated at grid
    q2block_per_nf :
        block coordinates
    xgrid :
        x grid
    """
    all_blocks = []

    # fake xfxQ2
    def pdf_xq2(pid, x, Q2):
        x_idx = xgrid.index(x)
        return x * evolved_PDF[(Q2, nf)][pid][x_idx]

    # loop on nf patches
    for nf, q2grid in q2block_per_nf.items():
        block = genpdf.generate_block(
            pdf_xq2,
            xgrid=xgrid,
            sorted_q2grid=q2grid,
            pids=np.array(br.flavor_basis_pids),
        )
        all_blocks.append(block)
    return all_blocks



"""Evolution Kernel Operators."""

from . import io, version
from .io.runcards import OperatorCard, TheoryCard
from .io.struct import EKO
from .runner import solve

__version__ = version.__version__

__all__ = [
    "io",
    "OperatorCard",
    "TheoryCard",
    "EKO",
    "solve",
]



"""EKO database configuration."""

from typing import Union

from eko.io.runcards import Legacy, OperatorCard, TheoryCard
from eko.io.types import RawCard


def update_runcards(
    theory: Union[RawCard, TheoryCard], operator: Union[RawCard, OperatorCard]
):
    """Update legacy runcards.

    This function is mainly defined for compatibility with the old interface.
    Prefer direct usage of :class:`Legacy` in new code.

    Consecutive applications of this function yield identical results::

        cards = update(theory, operator)
        assert update(*cards) == cards
    """
    if isinstance(theory, TheoryCard) or isinstance(operator, OperatorCard):
        # if one is not a dict, both have to be new cards
        assert isinstance(theory, TheoryCard)
        assert isinstance(operator, OperatorCard)
        return theory, operator

    cards = Legacy(theory, operator)
    return cards.new_theory, cards.new_operator



"""Recipes containing instructions for atomic computations."""

from typing import List

from ..io.items import Evolution, Matching, Recipe
from ..io.struct import EKO
from ..io.types import EvolutionPoint as EPoint
from ..matchings import Atlas, Segment
from . import commons


def _elements(ep: EPoint, atlas: Atlas) -> List[Recipe]:
    """Determine recipes to compute a given operator."""
    recipes = []

    #  expanded = eko.operator_card.configs.scvar_method is sv.Modes.expanded
    #  mu2f = mu2 * eko.theory_card.xif**2 if expanded else mu2

    blocks = atlas.matched_path(ep)
    for block in blocks:
        if isinstance(block, Segment):
            cliff = block.target in atlas.walls
            recipe = Evolution.from_atlas(block, cliff=cliff)
        else:
            recipe = Matching.from_atlas(block)

        recipes.append(recipe)

    return recipes


def _create(evolgrid: List[EPoint], atlas: Atlas) -> List[Recipe]:
    """Create all associated recipes."""
    recipes = []
    for ep in evolgrid:
        recipes.extend(_elements(ep, atlas))

    return list(set(recipes))


def create(eko: EKO):
    """Create and load all associated recipes."""
    atlas = commons.atlas(eko.theory_card, eko.operator_card)
    recs = _create(eko.operator_card.evolgrid, atlas)
    eko.load_recipes(recs)



"""Define possible scale variations schemes.

A Mathematica snippet to check the formulas is available in the extras
folder.
"""

import enum
from typing import Any, Dict

from ..io.types import ScaleVariationsMethod


class Modes(enum.IntEnum):
    """Enumerate scale variation modes."""

    unvaried = enum.auto()
    exponentiated = enum.auto()
    expanded = enum.auto()


def sv_mode(s: ScaleVariationsMethod) -> Modes:
    """Return the scale variation mode.

    Parameters
    ----------
    s :
        string representation

    Returns
    -------
    i :
        int representation
    """
    if s is not None:
        return Modes[s.value]
    return Modes.unvaried


class ScaleVariationModeMixin:
    """Mixin to cast scale variation mode."""

    config: Dict[str, Any]

    @property
    def sv_mode(self) -> Modes:
        """Return the scale variation mode."""
        return sv_mode(self.config["ModSV"])



r"""The unpolarized, space-like anomalous dimension
:math:`\gamma_{ns,v}^{(3)}`."""

import numba as nb

from ......harmonics import cache as c
from ......harmonics.log_functions import lm11m1, lm12m1, lm13m1
from .gnsm import gamma_nsm


@nb.njit(cache=True)
def gamma_nss(n, nf, cache, variation):
    r"""Compute the |N3LO| sea non-single anomalous dimension.

    The routine is taken from :cite:`Moch:2017uml`.

    The :math:`nf^2` part is a high-accuracy (0.1% or better) parametrization
    of the exact expression obtained in :cite:`Davies:2016jie`, see xpns3m.f

    The :math:`nf^1` part is an approximation based on the first 9 odd moments.
    The two sets spanning the error estimate are called via  IMOD = 1
    and  IMOD = 2.  Any other value of IMOD invokes their average.

    Parameters
    ----------
    n : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache : numpy.ndarray
        Harmonic sum cache
    variation : int
        |N3LO| anomalous dimension variation

    Returns
    -------
    complex
        |N3LO| sea non-singlet anomalous dimension :math:`\gamma_{ns,s}^{(3)}`
    """
    S1 = c.get(c.S1, cache, n)
    S2 = c.get(c.S2, cache, n)
    S3 = c.get(c.S3, cache, n)

    # nf^1: two approximations
    P3NSA11 = (
        2880 / n**7
        - 11672.4 / n**6
        + 12802.560000000001 / n**5
        - 7626.66 / n**4
        + 6593.2 / n**3
        - 3687.6 / n**2
        + 4989.2 / (1 + n)
        - 6596.93 / (2 + n)
        + 1607.73 / (3 + n)
        + 60.4 * lm12m1(n, S1, S2)
        + 4.685 * lm13m1(n, S1, S2, S3)
    )
    P3NSA12 = (
        -2880 / n**7
        + 4066.32 / n**6
        - 5682.24 / n**5
        + 5540.88 / n**4
        + 546.1 / n**3
        - 2987.83 / n**2
        + 2533.54 / n
        - 1502.75 / (1 + n)
        - 2297.56 / (2 + n)
        + 1266.77 / (3 + n)
        - 254.63 * lm11m1(n, S1)
        - 0.28953 * lm13m1(n, S1, S2, S3)
    )

    # nf^2 (parametrized)
    P3NSSA2 = (
        47.4074 / n**6
        - 142.222 / n**5
        + 32.1201 / n**4
        - 132.824 / n**3
        + 647.397 / n**2
        + 19.7 * lm11m1(n, S1)
        - 3.43547 * lm12m1(n, S1, S2)
        - 1262.0951538579698 / n
        - 187.17000000000002 / (1 + n) ** 4
        + 453.885 / (1 + n) ** 3
        + 147.01749999999998 / (1 + n) ** 2
        + 1614.1000000000001 / (1 + n)
        - 380.12500000000006 / (2 + n)
        - 42.575 / (3 + n)
        + (42.977500000000006 * S2) / n
        + (0.0900000000000047 * (477.52777777775293 + n) * S1) / (n**2 * (1 + n))
    )

    if variation == 1:
        P3NSSA = nf * P3NSA11 + nf**2 * P3NSSA2
    elif variation == 2:
        P3NSSA = nf * P3NSA12 + nf**2 * P3NSSA2
    else:
        P3NSSA = 0.5 * nf * (P3NSA11 + P3NSA12) + nf**2 * P3NSSA2
    return -P3NSSA


@nb.njit(cache=True)
def gamma_nsv(n, nf, cache, variation):
    r"""Compute the |N3LO| valence non-singlet anomalous dimension.

    Parameters
    ----------
    n : complex
        Mellin moment
    nf : int
        Number of active flavors
    cache: numpy.ndarray
        Harmonic sum cache
    variation : int
        |N3LO| anomalous dimension variation

    Returns
    -------
    complex
        |N3LO| valence non-singlet anomalous dimension
        :math:`\gamma_{ns,v}^{(3)}(N)`
    """
    return gamma_nsm(n, nf, cache, variation) + gamma_nss(n, nf, cache, variation)



"""|APFEL| interface."""

import time

import numpy as np
from banana.benchmark.external.apfel_utils import load_apfel

from eko import basis_rotation as br


def compute_apfel_data(
    theory, operators, pdf, skip_pdfs, rotate_to_evolution_basis=False
):
    """Run APFEL to compute operators.

    Parameters
    ----------
    theory : dict
        theory card
    operators : dict
        operators card
    pdf : lhapdf_type
        pdf
    skip_pdfs : list
        list of pdfs (pid or name) to skip
    rotate_to_evolution_basis: bool
        rotate to evolution basis

    Returns
    -------
    ref : dict
        output containing: target_xgrid, values
    """
    target_xgrid = operators["interpolation_xgrid"]
    pdf_name = pdf.set().name

    # Load apfel
    apf_start = time.perf_counter()
    if theory["ModEv"] in ["EXA", "perturbative-exact"]:
        theory["ModEv"] = "EXA"
    elif theory["ModEv"] in ["EXP", "decompose-expanded", "perturbative-expanded"]:
        theory["ModEv"] = "EXP"
    elif theory["ModEv"] in ["TRN", "ordered-truncated"]:
        theory["ModEv"] = "TRN"
    else:
        raise ValueError(f"Method {theory['ModEv']} is not recognized. ")
    apfel = load_apfel(theory, operators, pdf_name)

    # Truncated Epsilon
    # APFEL::SetEpsilonTruncation(1E-1);
    #
    # Set maximum scale
    # APFEL::SetQLimits(theory.Q0, theory.QM );
    #
    # if (theory.SIA)
    # {
    #   APFEL::SetPDFSet("kretzer");
    #   APFEL::SetTimeLikeEvolution(true);
    # }

    # Set APFEL interpolation grid
    #
    # apfel.SetNumberOfGrids(3)
    # apfel.SetGridParameters(1, 50, 3, 1e-5)
    # apfel.SetGridParameters(2, 50, 3, 2e-1)
    # apfel.SetGridParameters(3, 50, 3, 8e-1)

    # init evolution
    apfel.SetPolarizedEvolution(operators["polarized"])
    apfel.InitializeAPFEL()
    print(f"Loading APFEL took {(time.perf_counter() - apf_start)} s")

    # Run
    apf_tabs = {}
    for mu in operators["mugrid"]:
        apfel.EvolveAPFEL(theory["Q0"], mu)
        print(f"Executing APFEL took {(time.perf_counter() - apf_start)} s")

        tab = {}
        for pid in br.flavor_basis_pids:
            if pid in skip_pdfs:
                continue

            # collect APFEL
            apf = []
            for x in target_xgrid:
                if pid != 22:
                    xf = apfel.xPDF(pid if pid != 21 else 0, x)
                else:
                    xf = apfel.xgamma(x)
                # if pid == 4:
                #     print(pid,x,xf)
                apf.append(xf)
            tab[pid] = np.array(apf)

        # rotate if needed
        if rotate_to_evolution_basis:
            qed = theory["QED"] > 0
            if not qed:
                evol_basis = br.evol_basis
                rotate_flavor_to_evolution = br.rotate_flavor_to_evolution
            else:
                evol_basis = br.unified_evol_basis
                rotate_flavor_to_evolution = br.rotate_flavor_to_unified_evolution
            pdfs = np.array(
                [
                    tab[pid] if pid in tab else np.zeros(len(target_xgrid))
                    for pid in br.flavor_basis_pids
                ]
            )
            evol_pdf = rotate_flavor_to_evolution @ pdfs
            tab = dict(zip(evol_basis, evol_pdf))

        apf_tabs[mu**2] = tab

    ref = {
        "target_xgrid": target_xgrid,
        "values": apf_tabs,
    }

    return ref



"""Legacy interface to files created with v0.13.5.

Although the data version v1 was already before v0.13.5 we only support
that API version.
"""

from .paths import InternalPaths


def update_metadata(paths: InternalPaths, raw: dict) -> dict:
    """Modify the raw metadata to the new format.

    Parameters
    ----------
    paths:
        base paths to the EKO
    raw:
        raw yaml content

    Returns
    -------
    dict
        compatible raw yaml content
    """
    raw["data_version"] = 1
    raw["xgrid"] = raw["bases"]["xgrid"]
    del raw["bases"]

    return raw


def update_theory(raw: dict) -> dict:
    """Modify the raw theory card to the new format.

    Parameters
    ----------
    raw:
        raw yaml content

    Returns
    -------
    dict
        compatible raw yaml content
    """
    raw["couplings"]["ref"] = (
        raw["couplings"]["scale"],
        raw["couplings"]["num_flavs_ref"],
    )
    raw["matching_order"] = [0, 0]
    # adjust couplings
    for key in ["num_flavs_ref", "max_num_flavs", "scale"]:
        del raw["couplings"][key]
    # adjust heavy
    for key in ["intrinsic_flavors", "num_flavs_init", "num_flavs_max_pdf"]:
        del raw["heavy"][key]

    # update old names
    if "use_fhmv" in raw:
        raw["use_fhmruvv"] = raw["use_fhmv"]
        del raw["use_fhmv"]

    return raw


def update_operator(raw_op: dict, raw_th) -> dict:
    """Modify the raw operator card to the new format.

    Parameters
    ----------
    raw_op:
        raw operator yaml content
    raw_th:
        raw theory yaml content
    Returns
    -------
    dict
        compatible raw yaml content
    """
    raw_op["configs"]["n_integration_cores"] = 1
    raw_op["init"] = (raw_op["mu0"], raw_th["heavy"]["num_flavs_init"])
    del raw_op["mu0"]

    return raw_op



"""Legacy interface to files created with v0.14.

Although the data version v2 was never assigned to v0.14 we use it for
exactly that API version.
"""

from .paths import InternalPaths


def update_metadata(paths: InternalPaths, raw: dict) -> dict:
    """Modify the raw metadata to the new format.

    Parameters
    ----------
    paths:
        base paths to the EKO
    raw:
        raw yaml content

    Returns
    -------
    dict
        compatible raw yaml content
    """
    raw["data_version"] = 2
    raw["xgrid"] = raw["bases"]["xgrid"]
    del raw["bases"]
    return raw


def update_theory(raw: dict) -> dict:
    raw["couplings"]["ref"] = (
        raw["couplings"]["scale"],
        raw["couplings"]["num_flavs_ref"],
    )

    # adjust couplings
    for key in ["num_flavs_ref", "max_num_flavs", "scale"]:
        del raw["couplings"][key]
    # adjust heavy
    for key in ["intrinsic_flavors", "num_flavs_init", "num_flavs_max_pdf"]:
        del raw["heavy"][key]

    return raw


def update_operator(raw_op: dict, raw_th) -> dict:
    raw_op["init"] = (raw_op["mu0"], raw_th["heavy"]["num_flavs_init"])
    raw_op["mugrid"] = raw_op["mugrid"]
    del raw_op["mu0"]
    return raw_op
